"""
tests for saving arbors and trees



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.utilities.testing import \
    compare_trees, \
    requires_file, \
    TempDirTest

import ytree

CT = "consistent_trees/tree_0_0_0.dat"
TCT = "tiny_ctrees/locations.dat"

class SaveArborTest(TempDirTest):
    @requires_file(CT)
    def test_default_save(self):
        a = ytree.load(CT)
        fn = a.save_arbor(max_file_size=512)

        a2 = ytree.load(fn)
        assert a.size == a2.size

    @requires_file(CT)
    def test_save_non_roots(self):
        a = ytree.load(CT)

        my_trees = [list(a[0]["tree"])[1], list(a[1]["tree"])[1]]
        fn = a.save_arbor(trees=my_trees)
        a2 = ytree.load(fn)

        fields = a2.field_list[:]
        fields.remove("desc_uid")

        for t1, t2 in zip(a2, my_trees):
            compare_trees(t1, t2, fields=fields)

    @requires_file(CT)
    def test_save_field_list(self):
        a = ytree.load(CT)

        my_trees = [list(a[0]["tree"])[1], list(a[1]["tree"])[1]]
        fn = a.save_arbor(trees=my_trees, fields=["mass", "redshift"])
        a2 = ytree.load(fn)

        assert len(a2.field_list) == 4
        assert sorted(["mass", "redshift", "uid", "desc_uid"]) == \
          sorted(a2.field_list)

        fields = a2.field_list[:]
        fields.remove("desc_uid")

        for t1, t2 in zip(a2, my_trees):
            compare_trees(t1, t2, fields=fields)

    @requires_file(CT)
    def test_save_tree(self):
        a = ytree.load(CT)

        t = a[0]
        fn = t.save_tree()
        a2 = ytree.load(fn)

        fields = a2.field_list[:]
        fields.remove("desc_uid")
        compare_trees(t, a2[0], fields=fields)

    @requires_file(CT)
    def test_save_tree_nonroot(self):
        a = ytree.load(CT)

        t = list(a[0]["tree"])[1]
        fn = t.save_tree()
        a2 = ytree.load(fn)

        fields = a2.field_list[:]
        fields.remove("desc_uid")
        compare_trees(t, a2[0], fields=fields)

    @requires_file(TCT)
    def test_save_in_place_true(self):
        a = ytree.load(TCT)

        fn = a.save_arbor()
        a2 = ytree.load(fn)
        a2.add_analysis_field("test_field", "", default=-1.)

        # this should be only a subset of all trees
        trees = list(a2[a2["mass"] > 3e11])
        for tree in trees:
            for node in tree["forest"]:
                node["test_field"] = 5

        fn2 = a2.save_arbor(trees=trees)
        a3 = ytree.load(fn2)

        # check if the arbor size is unchanged
        assert a3.size == a.size
        # check if number of trees with altered field is right
        assert (a3["test_field"] == 5).sum() == len(trees)

        trees = a3[a3["mass"] > 3e11]
        for tree in trees:
            assert (tree["forest", "test_field"] == 5).all()

    @requires_file(TCT)
    def test_save_in_place_false(self):
        a = ytree.load(TCT)

        fn = a.save_arbor()
        a2 = ytree.load(fn)
        a2.add_analysis_field("test_field", "", default=-1.)

        # this should be only a subset of all trees
        trees = list(a2[a2["mass"] > 3e11])
        for tree in trees:
            for node in tree["forest"]:
                node["test_field"] = 5

        fn2 = a2.save_arbor(trees=trees, save_in_place=False)
        a3 = ytree.load(fn2)

        # check if the arbor size changed (it should be!)
        assert a3.size == len(trees)
        # check if number of trees with altered field is right
        assert (a3["test_field"] == 5).sum() == len(trees)

        trees = a3[:]
        for tree in trees:
            assert (tree["forest", "test_field"] == 5).all()

    @requires_file(TCT)
    def test_save_roots_only(self):
        a = ytree.load(TCT)

        fn = a.save_arbor()
        a2 = ytree.load(fn)
        a2.add_analysis_field("test_field", "", default=-1.)

        # this should be only a subset of all trees
        trees = list(a2[a2["mass"] > 3e11])
        for tree in trees:
            for node in tree["forest"]:
                node["test_field"] = 5

        fn2 = a2.save_arbor(trees=trees, save_roots_only=True)
        a3 = ytree.load(fn2)

        # check if the arbor size is unchanged
        assert a3.size == a.size
        # check if number of trees with altered field is right
        assert (a3["test_field"] == 5).sum() == len(trees)

        trees = a3[a3["mass"] > 3e11]
        for tree in trees:
            # the root should be changed, but the rest of the tree should not
            assert tree["test_field"] == 5
            assert (tree["forest", "test_field"][1:] == -1).all()

    @requires_file(TCT)
    def test_save_in_place_roots_only(self):
        a = ytree.load(TCT)

        fn = a.save_arbor()
        a2 = ytree.load(fn)
        a2.add_analysis_field("test_field", "", default=-1.)

        trees = list(a2[a2["mass"] > 3e11])
        with self.assertRaises(ValueError):
            a2.save_arbor(trees=trees, save_in_place=False,
                          save_roots_only=True)



"""
ytree logger



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import logging
import sys

# CRITICAL 50
# ERROR    40
# WARNING  30
# INFO     20
# DEBUG    10
# NOTSET   10

ufstring = "%(name)-3s: [%(levelname)-9s] %(asctime)s %(message)s"

ytreeLogger = logging.getLogger("ytree")
ytree_sh = logging.StreamHandler(stream=sys.stderr)
formatter = logging.Formatter(ufstring)
ytree_sh.setFormatter(formatter)
ytreeLogger.addHandler(ytree_sh)
ytreeLogger.setLevel(20)
ytreeLogger.propagate = False

def set_parallel_logger(comm):
    if comm.size == 1: return
    f = logging.Formatter(f"P{comm.rank:03d} {ufstring}")
    if len(ytreeLogger.handlers) > 0:
        ytreeLogger.handlers[0].setFormatter(f)

class fake_pbar:
    def __init__(self, *args):
        pass
    def update(self, *args):
        pass
    def finish(self):
        pass

class log_level():
    """
    Context manager for setting log level.
    """
    def __init__(self, minlevel, mylog=None):
        if mylog is None:
            mylog = ytreeLogger
        self.mylog = mylog
        self.minlevel = minlevel
        self.level = mylog.level

    def __enter__(self):
        if self.level > 10 and self.level < self.minlevel:
            self.mylog.setLevel(40)

    def __exit__(self, *args):
        self.mylog.setLevel(self.level)



"""
testing utilities



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
from numpy.testing import \
    assert_equal, \
    assert_almost_equal, \
    assert_array_equal
import os
import pytest
import shutil
import subprocess
import sys
import tempfile
from unittest import \
    skipIf, TestCase
from yt.funcs import \
    get_pbar

from numpy.dtypes import StringDType

from ytree.data_structures.load import load
from ytree.frontends.ytree import YTreeArbor
from ytree.utilities.io import dirname
from ytree.utilities.loading import check_path, get_path
from ytree.utilities.logger import ytreeLogger as mylog

try:
    from mpi4py import MPI
except ModuleNotFoundError:
    MPI = None

generate_results = \
  int(os.environ.get("YTREE_GENERATE_TEST_RESULTS", 0)) == 1

def requires_file(filename):

    def ffalse(func):
        return None

    def ftrue(func):
        return func

    if not isinstance(filename, list):
        filename = [filename]
    try:
        [get_path(fn) for fn in filename]
    except IOError:
        return ffalse
    return ftrue

class TempDirTest(TestCase):
    """
    A test class that runs in a temporary directory and
    removes it afterward.
    """

    def setUp(self):
        self.curdir = os.getcwd()
        self.tmpdir = tempfile.mkdtemp()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)

class ParallelTest:
    """
    Test class for ytree parallelism.
    """

    base_filename = "tiny_ctrees/locations.dat"
    test_filename = "test_arbor/test_arbor.h5"
    test_script = None
    ncores = 4

    def check_values(self, arbor, my_args):
        group = my_args[1]

        if "nodes" in group:
            increment = int(group.split("-")[1])
        else:
            increment = None

        assert_array_equal(arbor["test_field"], 2 * arbor["mass"])
        for tree in arbor:
            if increment is None:
                assert_array_equal(tree[group, "test_field"], 2 * tree[group, "mass"])

            else:
                all_inds = np.arange(tree.tree_size)
                my_inds = all_inds[::increment]
                select = np.isin(all_inds, my_inds)
                assert_array_equal(
                    tree["forest", "test_field"][select],
                    2 * tree["forest", "mass"][select])

                # the ones we didn't do should all be -1
                assert_array_equal(
                    tree["forest", "test_field"][~select],
                    -np.ones_like(tree["forest", "test_field"][~select]))

    @skipIf(MPI is None, "mpi4py not installed")
    @pytest.mark.parallel
    def test_parallel(self):

        for i, my_args in enumerate(self.arg_sets):
            with self.subTest(i=i):

                args = [self.test_script, self.base_filename, self.test_filename] + \
                    [str(arg) for arg in my_args]
                comm = MPI.COMM_SELF.Spawn(sys.executable, args=args, maxprocs=self.ncores)
                comm.Disconnect()

                test_arbor = load(self.test_filename)
                self.check_values(test_arbor, my_args)

def run_command(command, timeout=None):
    try:
        proc = subprocess.run(command, shell=True, timeout=timeout)
        if proc.returncode == 0:
            success = True
        else:
            success = False
    except subprocess.TimeoutExpired:
        print ("Process reached timeout of %d s. (%s)" % (timeout, command))
        success = False
    except KeyboardInterrupt:
        print ("Killed by keyboard interrupt!")
        success = False
    return success

class ExampleScriptTest:
    """
    Tests for the code examples.
    """

    script_filename = None
    input_filename = None
    timeout = 60
    output_files = ()
    ncores = 1

    @pytest.mark.parallel
    def test_example(self):
        if self.script_filename is None:
            return

        if self.input_filename is not None:
            try:
                check_path(self.input_filename)
            except IOError:
                self.skipTest("test file missing")

        source_dir = dirname(__file__, level=3)
        script_path = os.path.join(
            source_dir, "doc", "source", "examples", self.script_filename)

        if self.ncores > 1:
            if MPI is None:
                self.skipTest("mpi4py not installed")
            comm = MPI.COMM_SELF.Spawn(sys.executable, args=[script_path],
                                       maxprocs=self.ncores)
            comm.Disconnect()
        else:
            command = f"{sys.executable} {script_path}"
            assert run_command(command, timeout=self.timeout)

        for fn in self.output_files:
            assert os.path.exists(fn)

class ArborTest:
    """
    A battery of tests for all frontends.
    """

    arbor_type = None
    test_filename = None
    load_kwargs = None
    load_callback = None
    groups = ("tree", "prog")
    num_data_files = None
    tree_skip = 1
    custom_vector_fields = None

    _arbor = None
    @property
    def arbor(self):
        if self._arbor is None:
            try:
                self.test_filename = get_path(self.test_filename)
            except IOError:
                self.skipTest("test file missing")

            if self.load_kwargs is None:
                self.load_kwargs = {}

            self._arbor = load(self.test_filename, **self.load_kwargs)
            if self.load_callback is not None:
                self.load_callback(self._arbor)
        return self._arbor

    def test_arbor_type(self):
        assert isinstance(self.arbor, self.arbor_type)

    def test_data_files(self):
        if self.num_data_files is None:
            return
        assert_equal(
            len(self.arbor.data_files), self.num_data_files,
            err_msg=f'Incorrect number of data files for {self.arbor}.')

    def test_get_root_nodes(self):
        for my_tree in get_random_trees(self.arbor, 47457, 5):
            verify_get_root_nodes(my_tree)

    def test_get_roor_nodes_nonroot(self):
        my_tree = list(self.arbor[0].ancestors)[0]
        verify_get_root_nodes(my_tree)

    def test_get_leaf_nodes(self):
        for my_tree in get_random_trees(self.arbor, 41153, 5):
            verify_get_leaf_nodes(my_tree)

    def test_get_leaf_nodes_ungrown_nonroot(self):
        my_tree = list(self.arbor[0].ancestors)[0]
        verify_get_leaf_nodes(my_tree)

    def test_get_node(self):
        for my_tree in get_random_trees(self.arbor, 47988, 5):
            verify_get_node(my_tree)

            ihalos = np.arange(1, my_tree.tree_size)
            gen = np.random.default_rng(312)
            gen.shuffle(ihalos)
            for ihalo in ihalos[:3]:
                my_halo = my_tree.get_node("forest", ihalo)
                verify_get_node(my_halo)

    def test_get_node_ungrown_nonroot(self):
        my_tree = list(self.arbor[0].ancestors)[0]
        my_halo = my_tree.get_node("forest", 0)
        node_list = list(my_tree["forest"])
        assert_equal(my_halo.uid, node_list[0].uid)

    def test_reset_node(self):
        t = self.arbor[0]
        ts0 = len(list(t['tree']))
        f0 = dict((field, t['tree', field])
                  for field in ['uid', 'desc_uid'])

        assert self.arbor.is_setup(t)
        assert self.arbor.is_grown(t)

        self.arbor.reset_node(t)

        for attr in self.arbor._reset_attrs:
            assert getattr(t, attr) is None
        assert_equal(len(t.field_data), 0)
        assert not self.arbor.is_setup(t)
        assert not self.arbor.is_grown(t)

        assert_equal(
            len(list(t['tree'])), ts0,
            err_msg=f'Trees are not the same size after resetting for {self.arbor}.')

        for field in f0:
            assert_array_equal(
                t['tree', field], f0[field],
                err_msg=f"Tree field {field} not the same after resetting for {self.arbor}.")

    def test_reset_nonroot(self):
        t = self.arbor[0]
        node = list(t['tree'])[1]
        ts0 = len(list(node['tree']))
        f0 = dict((field, node['tree', field])
                  for field in ['uid', 'desc_uid'])

        self.arbor.reset_node(node)

        assert_equal(
            len(list(node['tree'])), ts0,
            err_msg=f'Trees are not the same size after resetting for {self.arbor}.')

        for field in f0:
            assert_array_equal(
                node['tree', field], f0[field],
                err_msg=f"Tree field {field} not the same after resetting for {self.arbor}.")

    def test_save_and_reload(self):
        skip = self.tree_skip
        if skip > 1:
            trees = list(self.arbor[::skip])
        else:
            trees = None

        fn = self.arbor.save_arbor(trees=trees)

        save_arbor = load(fn)
        if self.load_callback is not None:
            self.load_callback(save_arbor)

        assert isinstance(save_arbor, YTreeArbor)
        compare_arbors(save_arbor, self.arbor, groups=self.groups, skip2=skip)

    def test_vector_fields(self):
        a = self.arbor
        t = a[0]

        if self.custom_vector_fields is not None:
            for vfield, cfields in self.custom_vector_fields:
                a.add_vector_field(vfield, vector_components=cfields)

        for field in a.field_info.vector_fields:
            mylog.info(f"Comparing vector field: {field}.")
            magfield = np.sqrt((a[field]**2).sum(axis=1))
            assert_array_equal(a[f"{field}_magnitude"], magfield,
                               err_msg=f"Magnitude field incorrect: {field}.")

            cfields = a.field_info[field]["vector_components"]

            for i, cfield in enumerate(cfields):
                assert_array_equal(
                    a[cfield], a[field][:, i],
                    err_msg=(f"Arbor vector field {field} does not match "
                             f"in dimension {i}."))

                assert_array_equal(
                    t[cfield], t[field][i],
                    err_msg=(f"Tree vector field {field} does not match "
                             f"in dimension {i}."))

                for group in ["prog", "tree"]:
                    assert_array_equal(
                        t[group, cfield], t[group, field][:, i],
                        err_msg=(f"{group} vector field {field} does not match "
                                 f"in dimension {i}."))

def get_tree_split(arbor):
    """
    Get a few separate ancestors from a tree.
    """
    ancs = None
    for tree in arbor:
        for halo in tree["tree"]:
            ancs = list(halo.ancestors)
            if len(ancs) > 1:
                break

    if ancs is None:
        raise RuntimeError("Could not find nodes to test with.")
    return ancs

def get_random_trees(arbor, seed, n):
    """
    Get n random trees from the arbor.
    """

    gen = np.random.default_rng(708)
    itrees = np.arange(arbor.size)
    gen.shuffle(itrees)
    for itree in itrees[:5]:
        yield arbor[itree]

def get_stringsafe_compare_arrays(arr1, arr2):
    """
    Convert to stringy arrays to a common type for comparison.
    """

    if StringDType() in (arr1.dtype, arr2.dtype):
        # Python 3.9 and earlier require this. Otherwise, you end up
        # with  "b'Alexander'" instead of 'Alexander'.
        if arr1.dtype != StringDType():
            arr1 = arr1.astype(str)
        arr1 = arr1.astype(StringDType())
        if arr2.dtype != StringDType():
            arr2 = arr2.astype(str)
        arr2 = arr2.astype(StringDType())

    return arr1, arr2

def compare_arbors(a1, a2, groups=None, fields=None, skip1=1, skip2=1):
    """
    Compare all fields for all trees in two arbors.
    """

    if groups is None:
        groups = ["tree", "prog"]

    if fields is None:
        fields = a1.field_list

    for i, field in enumerate(fields):
        c1, c2 = get_stringsafe_compare_arrays(
            a1[field][::skip1], a2[field][::skip2])
        mylog.info(f"Comparing arbor field: {field} ({i+1}/{len(fields)}).")
        assert_array_equal(c1, c2,
                           err_msg=f"Arbor field mismatch: {a1, a2, field}.")

    trees1 = list(a1[::skip1])
    trees2 = list(a2[::skip2])

    ntot = len(trees1)
    pbar = get_pbar("Comparing trees", ntot)
    for i, (t1, t2) in enumerate(zip(trees1, trees2)):
        compare_trees(t1, t2, groups=groups, fields=fields)
        pbar.update(i+1)
    pbar.finish()

def compare_trees(t1, t2, groups=None, fields=None):
    """
    Compare all fields between two trees.
    """

    if groups is None:
        groups = ["tree", "prog"]

    if fields is None:
        fields = t1.arbor.field_list

    for field in fields:
        for group in groups:
            c1, c2 = get_stringsafe_compare_arrays(
                t1[group, field], t2[group, field])
            assert_array_equal(
                c1, c2,
                err_msg=f"Tree comparison failed for {group} field: {field}.")
    t1.arbor.reset_node(t1)
    t2.arbor.reset_node(t2)

def compare_hdf5(fh1, fh2, compare=None, compare_groups=True,
                 **kwargs):
    """
    Compare all datasets between two hdf5 files.
    """

    if compare is None:
        compare = assert_array_equal
    if not isinstance(fh1, h5py.Group):
        fh1 = h5py.File(fh1, "r")
    if not isinstance(fh2, h5py.Group):
        fh2 = h5py.File(fh2, "r")

    if compare_groups:
        err_msg = f"{fh1.file.filename} and {fh2.file.filename} have different datasets in group {fh1.name}."
        assert_equal(sorted(list(fh1.keys())), sorted(list(fh2.keys())), err_msg=err_msg)

    for key in fh1.keys():
        if isinstance(fh1[key], h5py.Group):
            compare_hdf5(fh1[key], fh2[key],
                         compare_groups=compare_groups,
                         compare=compare, **kwargs)
        else:
            err_msg = f"{key} field not equal for {fh1.file.filename} and {fh2.file.filename}"
            if fh1[key].dtype == "int":
                assert_array_equal(fh1[key][()], fh2[key][()],
                                   err_msg=err_msg)
            else:
                compare(fh1[key][()], fh2[key][()],
                        err_msg=err_msg, **kwargs)

def assert_rel_equal(a1, a2, decimals, err_msg="", verbose=True):
    # We have nan checks in here because occasionally we have fields that get
    # weighted without non-zero weights.  I'm looking at you, particle fields!
    if isinstance(a1, np.ndarray):
        assert a1.size == a2.size
        # Mask out NaNs
        assert (np.isnan(a1) == np.isnan(a2)).all()
        a1[np.isnan(a1)] = 1.0
        a2[np.isnan(a2)] = 1.0
        # Mask out 0
        ind1 = np.array(np.abs(a1) < np.finfo(a1.dtype).eps)
        ind2 = np.array(np.abs(a2) < np.finfo(a2.dtype).eps)
        assert (ind1 == ind2).all()
        a1[ind1] = 1.0
        a2[ind2] = 1.0
    elif np.any(np.isnan(a1)) and np.any(np.isnan(a2)):
        return True
    if not isinstance(a1, np.ndarray) and a1 == a2 == 0.0:
        # NANS!
        a1 = a2 = 1.0
    return assert_almost_equal(
        np.array(a1) / np.array(a2), 1.0, decimals, err_msg=err_msg, verbose=verbose
    )

def assert_array_rel_equal(a1, a2, decimals=16, **kwargs):
    """
    Wraps assert_rel_equal with, but decimals is a keyword arg.
    """
    assert_rel_equal(a1, a2, decimals, **kwargs)

def verify_get_node(my_tree, n=3):
    """
    Unit tests for get_node.
    """
    for selector in ["forest", "tree", "prog"]:
        node_list = list(my_tree[selector])

        inodes = np.arange(len(node_list))
        gen = np.random.default_rng(847)
        gen.shuffle(inodes)

        for inode in inodes[:3]:
            my_node = my_tree.get_node(selector, inode)
            err_msg = f"get_node failed: {selector} " + \
              f"with {str(my_tree.arbor)}."
            assert_equal(my_node.uid, node_list[inode].uid, err_msg=err_msg)

            if selector == "forest":
                err_msg = "get_node index is not tree_id for " + \
                  f"{str(my_tree.arbor)}."
                assert_equal(my_node.tree_id, inode, err_msg=err_msg)

def verify_get_leaf_nodes(my_tree):
    """
    Unit tests for get_leaf_nodes.
    """
    for selector in ["forest", "tree", "prog"]:
        uids1 = np.array([node.uid for node in
                          my_tree.get_leaf_nodes(selector=selector)])
        uids2 = np.array([my_halo.uid for my_halo in my_tree[selector]
                          if not list(my_halo.ancestors)])

        err_msg=f"get_leaf_nodes failure for {selector} in {my_tree.arbor}."
        assert_equal(uids1, uids2, err_msg=err_msg)

def verify_get_root_nodes(my_tree):
    """
    Unit tests for get_root_nodes.
    """

    root_nodes1 = list(my_tree.get_root_nodes())
    for root_node in root_nodes1:
        assert_equal(
            root_node["desc_uid"], -1,
            err_msg=f"In {my_tree.arbor}: {root_node} has " + \
                    f"desc_uid={root_node['desc_uid']}, but expected -1.")

    root_nodes2 = [node for node in my_tree["forest"]
                    if node.descendent is None]

    uids1 = np.sort([node.uid for node in root_nodes1])
    uids2 = np.sort([node.uid for node in root_nodes2])

    assert_array_equal(
        uids1, uids2,
        err_msg=f"get_root_nodes failure in {my_tree.arbor}.")



from ytree.frontends.rockstar import \
    RockstarArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class RockstarArborTest(TempDirTest, ArborTest):
    arbor_type = RockstarArbor
    test_filename = "rockstar/rockstar_halos/out_0.list"
    num_data_files = 65
    tree_skip = 10



"""
parallel_tree_nodes test script



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from mpi4py import MPI
import numpy as np
import sys
import yt
yt.enable_parallelism()
import ytree

def run():
    input_fn, output_fn, selection, group = sys.argv[1:5]
    njobs = int(sys.argv[5])
    dynamic = bool(int(sys.argv[6]))

    # test the nodes kwarg
    # in this case, group will be something like "nodes-10", indicating that
    # we will do every 10th node in the tree.
    if "nodes" in group:
        increment = int(group.split("-")[1])
    else:
        increment = None

    a = ytree.load(input_fn)
    if "test_field" not in a.field_list:
        a.add_analysis_field("test_field", default=-1, units="Msun")

    trees = list(a[:8])

    for tree in trees:
        if increment is None:
            nodes = None
        else:
            inds = np.arange(0, tree.tree_size, increment)
            nodes = [tree.get_node("forest", i) for i in inds]

        for node in ytree.parallel_tree_nodes(
                tree,
                group=group,
                nodes=nodes,
                njobs=njobs,
                dynamic=dynamic):
            root = node.root
            yt.mylog.info(f"Doing {node.tree_id}/{root.tree_size} of {root._arbor_index}")
            node["test_field"] = 2 * node["mass"]

    if yt.is_root():
        a.save_arbor(filename=output_fn, trees=trees)


if __name__ == "__main__":
    comm = MPI.Comm.Get_parent()
    try:
        run()
    except BaseException:
        pass
    comm.Disconnect()



"""
Field detection classes



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import numpy as np

from ytree.utilities.exceptions import \
    ArborFieldNotFound, \
    ArborFieldDependencyNotFound

_selectors = ("forest", "tree", "prog")

class Detector(defaultdict):
    """
    Base class for detecting field dependencies and testing operations.
    """

    def __missing__(self, key):
        self._validate_key(key)
        self._generate_data(key)
        return self[key]

    def _validate_key(self, key):
        raise NotImplementedError

    def _generate_data(self, key):
        raise NotImplementedError

class FieldDetector(Detector):
    """
    A fake field data container used to calculate dependencies.
    """
    def __init__(self, arbor, name=None):
        self.arbor = arbor
        self.name = name

    def _validate_key(self, key):
        if key not in self.arbor.field_info:
            raise ArborFieldDependencyNotFound(self.name, key, self.arbor)

    def _generate_data(self, key):
        fi = self.arbor.field_info[key]
        if cfields:= fi.get("vector_components", False):
            data = np.ones((1, len(cfields)))
        else:
            data = np.ones(1)
        units = fi.get("units", "")
        self[key] = self.arbor.arr(data, units)

class SelectionDetector(Detector):
    """
    A TreeNode-like object to test select_halos criteria.
    """
    def __init__(self, arbor):
        self.arbor = arbor
        self.selectors = []

    def _validate_key(self, key):
        if not isinstance(key, tuple) and len(key) != 2:
            raise ValueError(f"Invalid selection criteria: {key}.")

        selector, field = key

        if selector not in _selectors:
            raise ValueError(f"Selector must be one of {_selectors}: {selector}.")

        if field not in self.arbor.field_info:
            raise ArborFieldNotFound(field, self.arbor)

    def _generate_data(self, key):
        selector, field = key
        if selector not in self.selectors:
            self.selectors.append(selector)

        fi = self.arbor.field_info[field]
        units = fi.get("units", "")
        self[key] = self.arbor.arr(np.ones(1), units)



"""
CSVArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.data_structures.io import CatalogDataFile

class CSVDataFile(CatalogDataFile):
    def open(self):
        self.fh = open(self.filename, mode="r")

    def _parse_header(self):
        pass

    def _read_data_select(self, rfields, tree_nodes, dtypes):
        if not rfields:
            return {}

        # We want to support nodes with missing descendents,
        # but this requires we manually reset their desc_uid to -1.
        reset_desc_uid = "desc_uid" in rfields

        fi = self.arbor.field_info
        nt = len(tree_nodes)
        field_data = \
          self._create_field_arrays(rfields, dtypes, size=nt)

        self.open()
        f = self.fh
        sep = self.arbor.sep
        for i, tree_node in enumerate(tree_nodes):
            f.seek(tree_node._offset)
            line = f.readline()
            sline = line.split(sep)
            for field in rfields:
                field_data[field][i] = sline[fi[field]["column"]]

            if reset_desc_uid and tree_node.is_root:
                field_data["desc_uid"][i] = -1
        self.close()

        return field_data



from ytree.frontends.ahf import \
    AHFCRMArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class AHFCRMArborTest(TempDirTest, ArborTest):
    arbor_type = AHFCRMArbor
    test_filename = "AHF_100_tiny/GIZMO-NewMDCLUSTER_0047.snap_128.parameter"
    num_data_files = 5
    tree_skip = 100



"""
Gadget4Arbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
import re

from yt.funcs import \
    get_pbar

from ytree.data_structures.arbor import \
    SegmentedArbor

from ytree.frontends.gadget4.fields import \
    Gadget4FieldInfo
from ytree.frontends.gadget4.io import \
    Gadget4DataFile, \
    Gadget4TreeFieldIO

class Gadget4Arbor(SegmentedArbor):
    """
    Gadget4 inline merger tree format.
    """

    _suffix = ".hdf5"
    _data_file_class = Gadget4DataFile
    _field_info_class = Gadget4FieldInfo
    _tree_field_io_class = Gadget4TreeFieldIO
    _node_io_attrs = ("_fi", "_si", "_fei", "_ei")

    def _get_data_files(self):
        if self._nfiles == 1:
            files = [self.parameter_filename]
        else:
            suffix = self._suffix
            reg = re.search(rf"^(.+\D)\d+{suffix}$", self.parameter_filename)
            if reg is None:
                raise RuntimeError(
                    f"Cannot determine numbering system for {self.filename}.")
            prefix = reg.groups()[0]
            files = [f"{prefix}{i}{suffix}" for i in range(self._nfiles)]

        self.data_files = [self._data_file_class(f) for f in files]

    def _parse_parameter_file(self):
        f = h5py.File(self.parameter_filename, mode="r")

        g = f["Header"]
        self._size = int(g.attrs["Ntrees_Total"])
        self._nfiles = int(g.attrs["NumFiles"])

        g = f["Parameters"]
        self.hubble_constant = g.attrs["HubbleParam"]
        self.omega_matter = g.attrs["Omega0"]
        self.omega_lambda = g.attrs["OmegaLambda"]
        self.box_size = self.quan(g.attrs["BoxSize"], "Mpc/h")
        
        self._redshifts = f["TreeTimes/Redshift"][()]
        self._times = f["TreeTimes/Time"][()]

        field_list = []
        if self._size == 0:
            field_list = []
            return

        g = f["TreeHalos"]
        for d in g:
            dshape = g[d].shape
            if len(dshape) == 1:
                field_list.append(d)
            else:
                field_list.extend(
                    [f"{d}_{i}" for i in range(dshape[1])])

        self.field_list = field_list
        fi = dict((field, {}) for field in field_list)
        for field in ["uid", "desc_uid"]:
            self.field_list.append(field)
            fi[field] = {"units": "", "source": "arbor"}
        fi["desc_uid"]["dependencies"] = ["TreeDescendant"]
        self.field_info.update(fi)

    def _plant_trees(self):
        if self.is_planted or self._size == 0:
            return

        istart = 0
        file_sizes = np.empty(len(self.data_files), dtype=int)
        offset = np.empty(self._size, dtype=int)
        pbar = get_pbar("Planting trees", self._size)
        for idf, data_file in enumerate(self.data_files):
            file_sizes[idf] = data_file.nnodes
            ntrees = data_file.ntrees
            iend = istart + ntrees
            my_slice = slice(istart, iend)

            if ntrees > 0:
                self._node_info["_fi"][my_slice] = idf
                self._node_info["_tree_size"][my_slice] = data_file.tree_sizes
                offset[my_slice] = data_file.offsets

            istart += ntrees
            pbar.update(iend)

        file_end_offset = file_sizes.cumsum()
        file_start_offset = file_end_offset - file_sizes
        tree_size = self._node_info["_tree_size"]
        file_start_index = np.digitize(offset, file_end_offset)
        file_end_index = np.digitize(offset+tree_size, file_end_offset, right=True)

        self._node_info["_fi"] = file_start_index
        self._node_info["_fei"] = file_end_index
        self._node_info["_si"] = offset - file_start_offset[file_start_index]
        self._node_info["_ei"] = offset + tree_size - file_start_offset[file_end_index]
        self._node_info["uid"] = offset
        pbar.finish()

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Should be an hdf5 file with a few key attributes.
        """
        fn = args[0]

        if not h5py.is_hdf5(fn):
            return False

        attrs = ["Nhalos_ThisFile", "Nhalos_Total",
                 "Ntrees_ThisFile", "Ntrees_Total",
                 "NumFiles"]
        groups = ["TreeHalos", "TreeTable", "TreeTimes"]

        with h5py.File(fn, mode="r") as f:
            g = f["Header"]
            for attr in attrs:
                if attr not in g.attrs:
                    return False
            for group in groups:
                if group not in f:
                    return False
        return True



"""
NodeLink class



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

class NodeLink:
    __slots__ = ('tree_id', 'descendent', 'ancestors')

    def __init__(self, tree_id):
        self.tree_id = tree_id
        self.descendent = None
        self.ancestors = []

    def add_ancestor(self, node):
        self.ancestors.append(node)
        node.descendent = self



"""
parallel_trees test script



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from mpi4py import MPI
import sys
import yt
yt.enable_parallelism()
from ytree.utilities.testing import get_tree_split
import ytree

def run():
    input_fn, output_fn, selection, group = sys.argv[1:5]
    njobs = int(sys.argv[5])
    dynamic = bool(int(sys.argv[6]))
    if len(sys.argv) > 7:
        save_every = int(sys.argv[7])
    else:
        save_every = None

    a = ytree.load(input_fn)
    if "test_field" not in a.field_list:
        a.add_analysis_field("test_field", default=-1, units="Msun")

    if selection == "all":
        trees = list(a[:8])
    elif selection == "nonroot":
        trees = get_tree_split(a)
    else:
        print (f"Bad selection: {selection}.")
        sys.exit(1)

    for tree in ytree.parallel_trees(trees, njobs=njobs, dynamic=dynamic,
                                     save_every=save_every, filename=output_fn):
        for node in tree[group]:
            root = node.root
            yt.mylog.info(f"Doing {node.tree_id}/{root.tree_size} of {root._arbor_index}")
            node["test_field"] = 2 * node["mass"]


if __name__ == "__main__":
    comm = MPI.Comm.Get_parent()
    try:
        run()
    except BaseException:
        pass
    comm.Disconnect()



"""
tests for select_halos function



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from numpy.testing import assert_raises

import ytree

from ytree.utilities.exceptions import ArborFieldNotFound
from ytree.utilities.testing import requires_file

CT = "consistent_trees/tree_0_0_0.dat"

@requires_file(CT)
def test_select_halos():
    a = ytree.load(CT)

    halos = a.select_halos('tree["tree", "Orig_halo_ID"] == 0')
    hids = a.arr([h["Orig_halo_ID"] for h in halos])
    assert (hids == 0).all()

    halos = a.select_halos('tree["prog", "Orig_halo_ID"] == 0')
    hids = a.arr([h["Orig_halo_ID"] for h in halos])
    assert (hids == 0).all()

@requires_file(CT)
def test_select_halos_bad_input():
    a = ytree.load(CT)

    with assert_raises(ArborFieldNotFound):
        list(a.select_halos("(tree['forest', 'not_a_field'].to('Msun') > 1e13)"))

    with assert_raises(ValueError):
        list(a.select_halos("(tree['flock', 'mass'].to('Msun') > 1e13)"))
        list(a.select_halos("(tree['forest', 'mass'].to('Msun') > 1e13) & (tree['tree', 'redshift'] < 0.5)"))



"""
MoriaArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
import re

from ytree.data_structures.io import \
    DefaultRootFieldIO, \
    DataFile, \
    TreeFieldIO

class MoriaDataFile(DataFile):
    field_cache = None
    full_read = False
    fh = None

    def open(self):
        self.fh = h5py.File(self.filename, mode="r")

    def close(self):
        self.fh.close()
        self.fh = None

    def read_data(self, field, index):
        if self.full_read:
            if self.field_cache is None:
                self.field_cache = {}

            if field not in self.field_cache:
                self.field_cache[field] = self.fh[field][()]
            return self.field_cache[field][index]
        else:
            return self.fh[field][index]

class MoriaTreeFieldIO(TreeFieldIO):

    def get_fields(self, data_object, fields=None, **kwargs):
        """
        Call _setup_tree if asking for desc_uid so we can correct it.
        """

        if fields is not None and "desc_uid" in fields:
            self.arbor._setup_tree(data_object)
        super().get_fields(data_object, fields=fields, **kwargs)

    def _read_fields(self, root_node, fields, dtypes=None,
                     root_only=False):
        """
        Read fields from disk for a single tree.
        """

        fi = self.arbor.field_info
        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        rfields = list(set(fields).difference(afields))

        for afield in afields:
            rfields.extend(
                [dfield for dfield in fi[afield].get("dependencies", [])
                 if dfield not in rfields])

        data_file = self.arbor.data_files[0]
        close = False
        if data_file.fh is None:
            close = True
            data_file.open()

        if root_only:
            index = (-1, slice(root_node._si, root_node._si+1))
            dfilter = None
        else:
            index = (slice(None), slice(root_node._si, root_node._ei))
            if not hasattr(root_node, "_status"):
                status = data_file.read_data("status_sparta", index)
                status = self._transform_data(status)
                root_node._status = np.where(status != 0)[0]
            dfilter = root_node._status

        # this field cache is for temporarily storing vector field data
        field_cache = {}
        field_data = {}
        freg = re.compile(r"(^.+)_(\d+$)")
        for field in rfields:
            if fi[field].get("vector", False):
                fs = freg.search(field)
                fieldname, ifield = fs.groups()
                ifield = int(ifield)
                if fieldname not in field_cache:
                    field_cache[fieldname] = data_file.read_data(fieldname, index)
                data = field_cache[fieldname][..., ifield]
            else:
                data = data_file.read_data(field, index)
            field_data[field] = self._transform_data(
                data, my_filter=dfilter)

        if afields:
            field_data.update(self._get_arbor_fields(
                root_node, field_data, fields, afields, root_only,
                my_filter=dfilter))

        if close:
            data_file.close()

        self._apply_units(rfields, field_data)

        return field_data

    def _transform_data(self, data, my_filter=None):
        data = np.flip(data, axis=0).flatten()
        if my_filter is not None:
            data = data[my_filter]
        return data

    def _get_arbor_fields(self, root_node, field_data,
                          fields, afields, root_only,
                          my_filter=None):
        """
        Generate special fields from the arbor/treenode.
        """

        adata = {}

        if "snap_index" in fields:
            if root_only:
                adata["snap_index"] = \
                  np.array([self.arbor._redshifts.size-1], dtype=int)
            else:
                data, _ = np.mgrid[:self.arbor._redshifts.size,
                                   root_node._si:root_node._ei]
                adata["snap_index"] = self._transform_data(
                    data, my_filter=my_filter)

        return adata

class MoriaRootFieldIO(DefaultRootFieldIO):
    def _read_fields(self, storage_object, fields, dtypes=None):
        self.arbor._plant_trees()

        if dtypes is None:
            dtypes = {}

        fi = self.arbor.field_info
        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        rfields = list(set(fields).difference(afields))

        for afield in afields:
            rfields.extend(
                [dfield for dfield in fi[afield].get("dependencies", [])
                 if dfield not in rfields])

        data_file = self.arbor.data_files[0]
        data_file.open()
        fh = data_file.fh

        index = self.arbor._node_info['_si']
        field_cache = {}
        field_data = {}
        freg = re.compile(r"(^.+)_(\d+$)")
        for field in rfields:
            if fi[field].get("vector", False):
                fs = freg.search(field)
                fieldname, ifield = fs.groups()
                ifield = int(ifield)
                if fieldname not in field_cache:
                    field_cache[fieldname] = fh[fieldname][-1][index]
                data = field_cache[fieldname][:, ifield]
            else:
                data = fh[field][-1][index]

            dtype = dtypes.get(field)
            if dtype is not None:
                data = data.astype(dtype)
            field_data[field] = data

        self._apply_units(rfields, field_data)

        if afields:
            field_data.update(self._get_arbor_fields(
                field_data, fields, afields))

        data_file.close()

        return field_data

    def _get_arbor_fields(self, field_data, fields, afields):
        adata = {}

        if "snap_index" in fields:
            adata["snap_index"] = \
              np.full(self.arbor.size, self.arbor._redshifts.size-1)

        return adata



"""
Plot the most massive halo in the dataset.
"""

import ytree

a = ytree.load("consistent_trees/tree_0_0_0.dat")

imax = a["mass"].argmax()
my_tree = a[imax]
print(f"Most massive halo is {my_tree} with M = {my_tree['mass']}.")

p = ytree.TreePlot(my_tree)
p.min_mass_ratio = 0.001
p.save("most_massive.png")



"""
tests for plotting



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.utilities.testing import \
    requires_file, \
    TempDirTest

import ytree

CT = "consistent_trees/tree_0_0_0.dat"

class TreePlotTest(TempDirTest):
    @requires_file(CT)
    def test_default_plot(self):
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0])
        p.save()

    @requires_file(CT)
    def test_non_defaults(self):
        attrs = {'size_field': 'virial_radius',
                 'size_log': False,
                 'min_mass': 1e14,
                 'min_mass_ratio': 0.1}

        a = ytree.load(CT)

        for attr, val in attrs.items():
            p = ytree.TreePlot(a[0])
            setattr(p, attr, val)
            p.save()

    @requires_file(CT)
    def test_save(self):
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0])
        p.save('tree.png')

    @requires_file(CT)
    def test_dot_kwargs(self):
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0], dot_kwargs={'dpi': 200})
        p.save()

    @requires_file(CT)
    def test_node_function(self):
        def my_func(halo):
            label = f"{halo['uid']}"
            return {"label": label}
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0], node_function=my_func)
        p.save()

    @requires_file(CT)
    def test_node_function_bad(self):
        a = ytree.load(CT)
        with self.assertRaises(RuntimeError):
            ytree.TreePlot(a[0], node_function='notafunc')

    @requires_file(CT)
    def test_edge_function(self):
        def my_func(desc, anc):
            return {"color": "red"}
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0], edge_function=my_func)
        p.save()

    @requires_file(CT)
    def test_edge_function_bad(self):
        a = ytree.load(CT)
        with self.assertRaises(RuntimeError):
            ytree.TreePlot(a[0], edge_function='notafunc')



"""
load function



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.data_structures.arbor import \
    arbor_registry
from ytree.utilities.loading import \
    get_path

global load_warn
load_warn = True

def load(filename, method=None, **kwargs):
    """
    Load an Arbor, determine the type automatically.

    Parameters
    ----------
    filename : string
        Input filename.
    method : optional, string
        The type of Arbor to be loaded.  Existing types are:
        ConsistentTrees, Rockstar, TreeFarm, YTree.  If not
        given, the type will be determined based on characteristics
        of the input file.
    kwargs : optional, dict
        Additional keyword arguments are passed to _is_valid and
        the determined type.

    Returns
    -------
    Arbor

    Examples
    --------

    >>> import ytree
    >>> # saved Arbor (ytree format)
    >>> a = ytree.load("arbor/arbor.h5")
    >>> # Amiga Halo Finder
    >>> a = ytree.load("ahf_halos/snap_N64L16_000.parameter",
    ...                hubble_constant=0.7)
    >>> # consistent-trees
    >>> a = ytree.load("tiny_ctrees/locations.dat")
    >>> a = ytree.load("consistent_trees/tree_0_0_0.dat")
    >>> a = ytree.load("ctrees_hlists/hlists/hlist_0.12521.list")
    >>> # consistent-trees-hdf5
    >>> a = ytree.load("consistent_trees_hdf5/soa/forest.h5")
    >>> # LHaloTree
    >>> a = ytree.load("my_halos/trees_063.0")
    >>> # LHaloTree-hdf5
    >>> a = ytree.load("TNG50-4-Dark/trees_sf1_099.0.hdf5",
    ...                box_size=35, hubble_constant=0.6774,
    ...                omega_matter=0.3089, omega_lambda=0.6911)
    >>> # Moria
    >>> a = ytree.load("moria/moria_tree_testsim050.hdf5")
    >>> # Rockstar
    >>> a = ytree.load("rockstar_halos/out_0.list")
    >>> # treefarm
    >>> a = ytree.load("my_halos/fof_subhalo_tab_025.0.h5")
    >>> # TreeFrog
    >>> a = ytree.load("treefrog/VELOCIraptor.tree.t4.0-131.walkabletree.sage.forestID.foreststats.hdf5")

    """

    filename = get_path(filename)

    if method is None:
        candidates = []
        for candidate, c in arbor_registry.items():
            try:
                if c._is_valid(filename, **kwargs):
                    candidates.append(candidate)
            except BaseException:
                pass

        if len(candidates) == 0:
            raise IOError(f"Could not determine arbor type for {filename}.")
        elif len(candidates) > 1:
            errmsg = "Could not distinguish between these arbor types:\n"
            for candidate in candidates:
                errmsg += f"Possible: {candidate}.\n"
            errmsg += "Provide one of these types using the \"method\" keyword."
            raise IOError(errmsg)
        else:
            method = candidates[0]

    else:
        if method not in arbor_registry:
            raise IOError(
                f"Invalid method: {method}. Available: {arbor_registry.keys()}.")

    global load_warn
    if method not in ["YTree"] and load_warn:
        print(
f"""Additional features and improved performance (usually) by saving this arbor with \"save_arbor\" and reloading:
\t>>> a = ytree.load(\"{filename}\")
\t>>> fn = a.save_arbor()
\t>>> a = ytree.load(fn)"""
            )
        load_warn = False
    return arbor_registry[method](filename, **kwargs)



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--parallelonly", action="store_true", default=False,
        help="run only parallel tests"
    )
    parser.addoption(
        "--serialonly", action="store_true", default=False,
        help="run only serial tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "parallel: test uses parallelism")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--parallelonly"):
        skip_serial = pytest.mark.skip(reason="only running parallel tests")
        for item in items:
            if "parallel" not in item.keywords:
                item.add_marker(skip_serial)

    if config.getoption("--serialonly"):
        skip_parallel = pytest.mark.skip(reason="only running serial tests")
        for item in items:
            if "parallel" in item.keywords:
                item.add_marker(skip_parallel)



"""
parallel utilities



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.funcs import \
    get_pbar, \
    is_root as yt_is_root
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    _get_comm, \
    parallel_objects

from ytree.data_structures.load import load as ytree_load

def regenerate_node(arbor, node, new_index=None):
    """
    Regenerate the TreeNode using the provided arbor.

    This is to be used when the original arbor associated with the
    TreeNode no longer exists.

    If new_index is None, assume the arbor has the same structure
    as it did before. If new_index is not None, assume the node
    is now a root.
    """

    if new_index is None:
        root_node = node.find_root()
        new_node = root_node.get_node("forest", node.tree_id)
    else:
        root_node = arbor[new_index]
        new_node = root_node

    return new_node

def parallel_trees(trees, collect_results=True,
                   save_every=None, save_in_place=None,
                   save_roots_only=False, filename=None,
                   njobs=0, dynamic=False):
    """
    Iterate over a list of trees in parallel.

    Trees are divided up between the available processor groups. Analysis
    field values can then be assigned to halos within the tree. The trees
    will be saved either at the end of the loop or after a number of trees
    given by the ``save_every`` keyword are completed.

    This uses the yt
    :func:`~yt.utilities.parallel_tools.parallel_analysis_interface.parallel_objects`
    function, which is parallelized with MPI underneath and so is suitable
    for parallelism across compute nodes.

    Parameters
    ----------
    trees : list of :class:`~ytree.data_structures.tree_node.TreeNode` objects
        The trees to be iterated over in parallel.
    collect_results : optional, bool
        If True, then results stored in analysis fields will be collected
        by the root process. This must be set to True if saving is to be
        done. If False, results collection is ignored. This will result in
        a significant speedup. If you have no intention of altering analysis
        fields or do not need results to be recollected or saved, then this is
        the best option. Setting this to False will automatically set
        save_every to False as well.
        Default: True
    save_every : optional, int or False
        Number of trees to be completed before results are saved. This is used to
        save intermediate results in case scripts need to be restarted. This
        parameter results in different behavior depending on the value of the
        collect_results keyword. If save_every is set to:

            - integer: if collect_trees is True, the number of trees to complete
              before saving. If collect_trees is False, a ValueError exception will
              be raised.
            - False: no saving will be done. Results will still be collected if
              collect_results is True.
            - None: if collect_results if True, save will occur after iterating over
              all trees. If collect_results is False, no saving will be done.

        Default: None
    save_in_place : optional, bool or None
        If True, analysis fields will be saved to the original
        arbor, even if only a subset of all trees is provided
        with the trees keyword. If False and only a subset of
        all trees is provided, a new arbor will be created
        containing only the trees provided. If set to None,
        behavior is determined by the type of arbor loaded.
        If the arbor is a YTreeArbor (i.e., saved with
        save_arbor), save_in_place will be set to True. If
        not of this type, it will be set to False.
        Default: None
    save_roots_only : optional, bool
        If True, only field values of each node are saved.
        If False, field data for the entire tree stemming
        from that node are saved.
        Default:  False.
    filename : optional, string
        The name of the new arbor to be saved. If None, the naming convention
        will follow the filename keyword of the
        :func:`~ytree.data_structures.arbor.Arbor.save_arbor` function.
        Default: None
    njobs : optional, int
        The number of process groups for parallel iteration. Set to 0 to make
        the same number of process groups as available processors. Hence,
        each tree will be allocated to a single processor. Set to a number
        less than the total number of processors to create groups with multiple
        processors, which will allow for further parallelization within a tree.
        For example, running with 8 processors and setting njobs to 4 will result
        in 4 groups of 2 processors each.
        Default: 0
    dynamic : optional, bool
        Set to False to divide iterations evenly among process groups. Set to
        True to allocate iterations with a task queue. If True, the number of
        processors available will be one fewer than the total as one will act
        as the task queue server.
        Default: False

    Examples
    --------

    >>> import ytree
    >>> a = ytree.load("arbor/arbor.h5")
    >>> a.add_analysis_field("test_field", default=-1, units="Msun")
    >>> trees = list(a[:])
    >>> for tree in ytree.parallel_trees(trees):
    ...     for node in tree["forest"]:
    ...         node["test_field"] = 2 * node["mass"] # some analysis

    See Also
    --------
    parallel_tree_nodes, parallel_nodes

    """

    comm = _get_comm(())
    # This is the root process of the whole operation.
    # We may split into process groups later, in which case there
    # will be other root processes and we will uses calls to yt's
    # is_root to identify them.
    is_global_root = comm.comm is None or comm.comm.rank == 0

    if dynamic:
        if is_global_root:
            nt = len(trees)
        else:
            nt = None
        nt = comm.mpi_bcast(nt, root=0)
    else:
        nt = len(trees)

    if nt < 1:
        return

    arbor = trees[0].arbor
    afields = arbor.analysis_field_list

    if save_in_place is None:
        from ytree.frontends.ytree.arbor import YTreeArbor
        save_in_place = isinstance(arbor, YTreeArbor)

    # are we actually going to save anything?
    do_save = True
    if isinstance(save_every, (int, np.integer)):
        if collect_results is False:
            raise ValueError(
                "collect_results must be True if save_every is set to a number.")
    elif save_every is False:
        save_every = nt
        do_save = False
    elif save_every is None:
        do_save = collect_results
        save_every = nt
    nb = int(np.ceil(nt / save_every))

    for ib in range(nb):
        start = ib * save_every
        end = min(start + save_every, nt)

        my_items = range(start, end)

        arbor_storage = {}
        for tree_store, my_item in parallel_objects(
                my_items, storage=arbor_storage,
                njobs=njobs, dynamic=dynamic):

            my_tree = trees[my_item]
            yield my_tree

            # We use yt_is_root here because we want the root of this
            # workgroup running this iteration, not the global root.
            # It is this process's job to round up the results for this
            # iteration and place them in the storage object. This is the
            # the slowest part as we will copy field data for all of the
            # analysis fields for the entire forest or tree. In the future,
            # it may be worth trying to identify the nodes and analysis fields
            # which were actually altered and copying just them.
            if yt_is_root():
                if not collect_results:
                    continue

                # this is fast
                my_root = my_tree.find_root()

                tree_store.result_id = (my_root._arbor_index, my_tree.tree_id)

                # If the tree is not a root, only save the "tree" selection
                # as we could overwrite other trees in the forest.
                if save_roots_only:
                    pass
                elif my_tree.is_root:
                    selection = "forest"
                else:
                    selection = "tree"

                if save_roots_only:
                    tree_store.result = {field: my_tree[field]
                                         for field in afields}
                else:
                    # this is the slow part as we are grabbing all analysis
                    # fields for the entire tree and copying them.
                    tree_store.result = {field: my_tree[selection, field]
                                         for field in afields}

            else:
                tree_store.result_id = None

        # Use the global root to combine all results.
        # Both parts of this are fairly slow.
        if is_global_root and collect_results:
            my_trees = []
            pbar = get_pbar("Combining results", len(my_items))
            for i, my_item in enumerate(my_items):
                my_tree = trees[my_item]
                my_trees.append(my_tree)
                my_root = my_tree.find_root()
                key = (my_root._arbor_index, my_tree.tree_id)
                data = arbor_storage[key]

                if save_roots_only:
                    indices = my_tree.tree_id
                elif my_tree.is_root:
                    indices = slice(None)
                else:
                    indices = [my_tree._tree_field_indices]

                for field in afields:
                    if field not in my_root.field_data:
                        arbor._node_io._initialize_analysis_field(my_root, field)

                    my_root.field_data[field][indices] = data[field]
                pbar.update(i+1)
            pbar.finish()

            if do_save:
                if save_in_place:
                    save_trees = my_trees
                else:
                    save_trees = trees

                fn = arbor.save_arbor(filename=filename, trees=save_trees,
                                      save_in_place=save_in_place,
                                      save_roots_only=save_roots_only)
                new_arbor = ytree_load(fn)

                add_fields = set(arbor.derived_field_list).difference(
                    new_arbor.derived_field_list)
                for field in add_fields:
                    fi = arbor.field_info[field].copy()
                    ftype = fi.pop("type")
                    # skip aliases as they will have been saved as the field
                    if ftype == "alias":
                        continue

                    name = fi.pop("name")
                    function = fi.pop("function")
                    del fi["dependencies"]
                    new_arbor.add_derived_field(name, function, **fi)

                arbor = new_arbor

                trees = [regenerate_node(arbor, tree, new_index=i)
                         for i, tree in enumerate(trees)]

        comm.barrier()

def parallel_tree_nodes(tree, group="forest", nodes=None,
                        njobs=0, dynamic=False):
    """
    Iterate over nodes in a single tree in parallel.

    Nodes are divided up between the available processor groups. Analysis
    field values can then be assigned to each node (halo).

    Note, unlike the parallel_trees and parallel_nodes function, no saving
    is performed internally. Results saving with the
    :func:`~ytree.data_structures.arbor.Arbor.save_arbor` must be done
    manually.

    This uses the yt
    :func:`~yt.utilities.parallel_tools.parallel_analysis_interface.parallel_objects`
    function, which is parallelized with MPI underneath and so is suitable
    for parallelism across compute nodes.

    Parameters
    ----------
    tree : :class:`~ytree.data_structures.tree_node.TreeNode`
        The tree whose nodes will be iterated over.
    group : optional, str ("forest", "tree", or "prog")
        Determines the nodes to be iterated over in the tree: "forest" for
        all nodes in the forest, "tree" for all nodes in the tree, or "prog"
        for all nodes in the line of main progenitors.
        Default: "forest"
    nodes: optional, list
        A list of nodes to iterate over instead of using a forest, tree, or
        prog selection. If provided, this will supersede the value of the
        "group" keyword. Note, all nodes must be members of the tree given
        in the "tree" argument.
    njobs : optional, int
        The number of process groups for parallel iteration. Set to 0 to make
        the same number of process groups as available processors. Hence,
        each node will be allocated to a single processor. Set to a number
        less than the total number of processors to create groups with multiple
        processors, which will allow for further parallelization. For example,
        running with 8 processors and setting njobs to 4 will result in 4
        groups of 2 processors each.
        Default: 0
    dynamic : optional, bool
        Set to False to divide iterations evenly among process groups. Set to
        True to allocate iterations with a task queue. If True, the number of
        processors available will be one fewer than the total as one will act
        as the task queue server.
        Default: False

    Examples
    --------

    >>> import ytree
    >>> a = ytree.load("arbor/arbor.h5")
    >>> a.add_analysis_field("test_field", default=-1, units="Msun")
    >>> trees = list(a[:])
    >>> for tree in trees:
    ...     for node in ytree.parallel_tree_nodes(tree):
    ...         node["test_field"] = 2 * node["mass"] # some analysis

    See Also
    --------
    parallel_trees, parallel_nodes

    """

    afields = tree.arbor.analysis_field_list

    if nodes is None:
        my_halos = list(tree[group])
    else:
        my_halos = nodes

    tree_storage = {}
    for halo_store, ihalo in parallel_objects(
            range(len(my_halos)), storage=tree_storage,
            njobs=njobs, dynamic=dynamic):

        my_halo = my_halos[ihalo]
        yield my_halo
        if yt_is_root():
            halo_store.result_id = my_halo.tree_id
            halo_store.result = {field: my_halo[field]
                                 for field in afields}
        else:
            halo_store.result_id = -1

    # combine results for this tree
    if yt_is_root():
        for tree_id, result in sorted(tree_storage.items()):
            if tree_id == -1:
                continue
            my_halo = tree.get_node("forest", tree_id)

            for field, value in result.items():
                my_halo[field] = value

def parallel_nodes(trees, group="forest", collect_results=True,
                   save_every=None, save_in_place=None, filename=None,
                   njobs=None, dynamic=None):
    """
    Iterate over all nodes in a list of trees in parallel.

    Both trees and/or nodes within a tree are divided up between available
    process groups using multi-level parallelism. Analysis field values can
    then be assigned to all nodes (halos). Trees will be saved either at the
    end of the loop over all trees or after a number of trees given by the
    ``save_every`` keyword are completed.

    This uses the yt
    :func:`~yt.utilities.parallel_tools.parallel_analysis_interface.parallel_objects`
    function, which is parallelized with MPI underneath and so is suitable
    for parallelism across compute nodes.

    Parameters
    ----------
    trees : list of :class:`~ytree.data_structures.tree_node.TreeNode` objects
        The trees to be iterated over in parallel.
    group : optional, str ("forest", "tree", or "prog")
        Determines the nodes to be iterated over in the tree: "forest" for
        all nodes in the forest, "tree" for all nodes in the tree, or "prog"
        for all nodes in the line of main progenitors.
        Default: "forest"
    collect_results : optional, bool
        If True, then results stored in analysis fields will be collected
        by the root process. This must be set to True if saving is to be
        done. If False, results collection is ignored. This will result in
        a significant speedup. If you have no intention of altering analysis
        fields or do not need results to be recollected or saved, then this is
        the best option. Setting this to False will automatically set
        save_every to False as well.
        Default: True
    save_every : optional, int or False
        Number of trees to be completed before results are saved. This is used to
        save intermediate results in case scripts need to be restarted. This
        parameter results in different behavior depending on the value of the
        collect_results keyword. If save_every is set to:

            - integer: if collect_trees is True, the number of trees to complete
              before saving. If collect_trees is False, a ValueError exception will
              be raised.
            - False: no saving will be done. Results will still be collected if
              collect_results is True.
            - None: if collect_results if True, save will occur after iterating over
              all trees. If collect_results is False, no saving will be done.

        Default: None
    save_in_place : optional, bool or None
        If True, analysis fields will be saved to the original
        arbor, even if only a subset of all trees is provided
        with the trees keyword. This will essentially "update"
        the arbor in place. If False and only a subset of
        all trees is provided, a new arbor will be created
        containing only the trees provided. If set to None,
        behavior is determined by the type of arbor loaded.
        If the arbor is a YTreeArbor (i.e., saved with
        save_arbor), save_in_place will be set to True. If
        not of this type, it will be set to False.
        Default: None
    filename : optional, string
        The name of the new arbor to be saved. If None, the naming convention
        will follow the filename keyword of the
        :func:`~ytree.data_structures.arbor.Arbor.save_arbor` function.
        Default: None
    njobs : optional, tuple of ints
        The number of process groups for parallel iteration over trees and
        nodes within each tree. The first value sets behavior for iteration
        over trees and the second for iteration over nodes in a tree. For
        example, set to (1, 0) to parallelize only over nodes in a tree and
        (0, 1) to parallelize only over trees. For multi-level parallelism
        set the first value to a number less than the total number of
        processors and the second to 0. For example, if running with 8
        processors, set njobs to (2, 0) to iterate over each tree with a
        group of 4 processors. Within each tree, each of the 4 processors
        in the group will work on a single node. If set to None, njobs will
        be set to (0, 1) if there are most trees than processors (tree
        parallel) and (1, 0) otherwise (node parallel).
        Default: None
    dynamic : optional, tuples of bools
        Toggles task queue on/off for parallelism over trees (first value)
        and nodes within a tree (second). Set to a value False to divide
        iterations evenly among process groups. Set to True to allocate
        iterations with a task queue. If True, the number of
        processors available will be one fewer than the total as one will
        act as the task queue server. Yes, this can be set to (True, True).
        Try it.
        Default: (False, False)

    Examples
    --------

    >>> import ytree
    >>> a = ytree.load("arbor/arbor.h5")
    >>> a.add_analysis_field("test_field", default=-1, units="Msun")
    >>> trees = list(a[:])
    >>> for node in ytree.parallel_nodes(trees):
    ...     node["test_field"] = 2 * node["mass"] # some analysis

    See Also
    --------
    parallel_trees, parallel_tree_nodes

    """

    if njobs is None:
        comm = _get_comm(())
        # parallelize over trees if more trees than cores
        if len(trees) > comm.size:
            njobs = (0, 1)
        else:
            njobs = (1, 0)
    else:
        if not isinstance(njobs, (tuple, list)) or len(njobs) != 2:
            raise ValueError(f"njobs must be a tuple of length 2: {njobs}.")

    if dynamic is None:
        dynamic = (False, False)
    else:
        if not isinstance(dynamic, (tuple, list)) or len(dynamic) != 2:
            raise ValueError(f"dynamic must be a tuple of length 2: {dynamic}.")

    for tree in parallel_trees(
            trees, collect_results=collect_results,
            save_every=save_every, save_in_place=save_in_place,
            filename=filename, njobs=njobs[0], dynamic=dynamic[0]):

        for node in parallel_tree_nodes(
                tree, group=group,
                njobs=njobs[1], dynamic=dynamic[1]):

            yield node



from ytree.data_structures.load import load as ytree_load
from ytree.frontends.lhalotree_hdf5 import \
    LHaloTreeHDF5Arbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class LHaloTreeHDF5ArborTest(TempDirTest, ArborTest):
    arbor_type = LHaloTreeHDF5Arbor
    test_filename = "TNG50-4-Dark/trees_sf1_099.0.hdf5"
    load_kwargs = {"box_size": 35, "hubble_constant": 0.6774,
                   "omega_matter": 0.3089, "omega_lambda": 0.6911}
    groups = ("forest", "tree", "prog")
    num_data_files = 7
    tree_skip = 100

    @property
    def arbor(self):
        arbor = super().arbor
        if "mass" not in arbor.field_info:
            arbor.add_alias_field(
                "mass", "Group_M_TopHat200", units="Msun")
        return arbor

    def test_load_without_kwargs(self):
        filename = self.arbor.filename
        a = ytree_load(filename)
        a.save_arbor()



"""
tests for the plotting examples in the docs



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.utilities.testing import \
    requires_file, \
    TempDirTest

import ytree

def my_node(halo):
    prog = list(halo.find_root()['prog', 'uid'])
    if halo['uid'] in prog:
        color = 'red'
    else:
        color = 'black'

    label = \
    f"""
    id: {halo['uid']:d}
    mass: {halo['mass'].to('Msun'):.2e} Msun
    """

    my_kwargs = {"label": label, "fontsize": 8,
                 "shape": "square", "color": color}
    return my_kwargs

def my_edge(ancestor, descendent):
    if descendent['mass'] < ancestor['mass']:
        color = 'blue'
    elif descendent['mass'] / ancestor['mass'] > 10:
        color = 'green'
    else:
        color = 'black'

    my_kwargs = {"color": color, "penwidth": 5}
    return my_kwargs


AHF = "ahf_halos/snap_N64L16_000.parameter"

class TreePlotTest(TempDirTest):
    @requires_file(AHF)
    def test_tree_plot(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(a[0], dot_kwargs={'rankdir': 'LR', 'size': '"12,4"'})
        p.save('tree.png')

    @requires_file(AHF)
    def test_tree_plot_small(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(a[0], dot_kwargs={'rankdir': 'LR', 'size': '"12,4"'})
        p.min_mass_ratio = 0.01
        p.save('tree_small.png')

    @requires_file(AHF)
    def test_tree_plot_custom_node(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(a[0], dot_kwargs={'rankdir': "BT"},
                           node_function=my_node)
        p.save('tree_custom_node.png')

    @requires_file(AHF)
    def test_tree_plot_custom_edge(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(a[0], dot_kwargs={'rankdir': "BT"},
                           node_function=my_node,
                           edge_function=my_edge)
        p.save('tree_custom_edge.png')



"""
Plot the tree with the most halos.
"""

import ytree

a = ytree.load("consistent_trees/tree_0_0_0.dat")

tree_size = a.arr([t.tree_size for t in a])
imax = tree_size.argmax()
my_tree = a[imax]
print(f"Tree with most halos is {my_tree} with {my_tree.tree_size} halos.")

p = ytree.TreePlot(my_tree)
p.min_mass_ratio = 0.001
p.save("most_halos.png")



"""
API for ytree frontends



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import importlib

_frontends = [
    "ahf",
    "csv",
    "consistent_trees",
    "consistent_trees_hdf5",
    "gadget4",
    "lhalotree",
    "lhalotree_hdf5",
    "moria",
    "rockstar",
    "treefarm",
    "treefrog",
    "ytree",
]

class _frontend_container:
    def __init__(self):
        for frontend in _frontends:
            _mod = f"ytree.frontends.{frontend}"
            setattr(self, frontend, importlib.import_module(_mod))
        setattr(self, 'api', importlib.import_module('ytree.frontends.api'))
        setattr(self, '__name__', 'ytree.frontends.api')



"""
Calculate a halo's significance, defined as the time
integrated mass of all a halo's progenitors.

NOTE: this script includes extra code to make it run within
the test suite. To run conventionally, remove the following
lines and return the code block in the middle to the proper
tabbing (i.e., 4 spaces to the left).

    from mpi4py import MPI
    comm = MPI.Comm.Get_parent()
    try:

    except BaseException:
        pass
    comm.Disconnect()

"""

import yt
yt.enable_parallelism()
import ytree

def calc_significance(node):
   if node.descendent is None:
       dt = 0. * node["time"]
   else:
       dt = node.descendent["time"] - node["time"]

   sig = node["mass"] * dt
   if node.ancestors is not None:
       for anc in node.ancestors:
           sig += calc_significance(anc)

   node["significance"] = sig
   return sig


if __name__ == "__main__":
    # Remove the next three and final three lines to run conventionally.
    from mpi4py import MPI
    comm = MPI.Comm.Get_parent()
    try:

        a = ytree.load("tiny_ctrees/locations.dat")
        a.add_analysis_field("significance", "Msun*Myr")

        ap = ytree.AnalysisPipeline()
        ap.add_operation(calc_significance)

        trees = list(a[:])
        for tree in ytree.parallel_trees(trees, filename="halo_significance"):
            yt.mylog.info(f"Processing {tree}.")
            ap.process_target(tree)

        if yt.is_root():
            a2 = ytree.load("halo_significance/halo_significance.h5")
            a2.set_selector("max_field_value", "significance")
            prog = list(a2[0]["prog"])
            print (prog)

    # Remove the next three lines to run conventionally.
    except BaseException:
        pass
    comm.Disconnect()



"""
FieldIO class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import numpy as np
import os
import weakref

from ytree.utilities.exceptions import \
    ArborAnalysisFieldNotGenerated
from ytree.utilities.logger import \
    ytreeLogger as mylog

class FieldIO:
    """
    Base class for FieldIO classes.

    This object is resposible for field i/o for an Arbor.
    """

    def __init__(self, arbor, default_dtype=np.float64):
        self.arbor = weakref.proxy(arbor)
        self.default_dtype = default_dtype

    def _apply_units(self, fields, field_data):
        """
        Apply units to data that's just been read in.
        """

        fi = self.arbor.field_info
        for field in fields:
            units = fi[field].get("units", "")
            if units != "":
                field_data[field] = \
                  self.arbor.arr(field_data[field], units)

    def _initialize_analysis_field(self, storage_object, name):
        """
        Initialize an empty field array to be filled in later.
        """
        raise NotImplementedError

    def _determine_dtypes(self, fields, override_dict=None):
        """
        Figure out dtype for field.

        Priority is:
        1. override_dict
        2. self.arbor.field_info
        3. self.arbor.field_info._data_types
        4. self.default_dtype
        """
        if override_dict is None:
            override_dict = {}
        dtypes = override_dict.copy()

        fi = self.arbor.field_info
        fid = fi._data_types
        for field in fields:
            dtypes[field] = \
              dtypes.get(field, fi[field].get('dtype',
                fid.get(field, self.default_dtype)))
        return dtypes

    def _determine_field_storage(self, data_object):
        """
        Figure out which objects are responsible for storing field data.
        """
        return data_object

    def _read_fields(self, *args, **kwargs):
        """
        Read fields from disk.
        """
        raise NotImplementedError

    def _store_fields(self, storage_object, fields):
        """
        Only keep items on the fields list.
        """
        fcache = storage_object.field_data
        remove = set(fcache).difference(fields)
        for field in remove:
            del fcache[field]

    def get_fields(self, data_object, fields=None, **kwargs):
        """
        Load field data for a data object into storage structures.
        """

        if not fields:
            return

        # hack to make sure root_only is False if this is not a root
        if isinstance(self, TreeFieldIO) and \
          not data_object.is_root:
            kwargs["root_only"] = False

        storage_object = \
          self._determine_field_storage(data_object)
        fcache = storage_object.field_data

        fi = self.arbor.field_info

        # Determine size of field array we need.
        # Set to None if root_only since any size will do.
        if not hasattr(data_object, "root") or \
          kwargs.get("root_only", False):
            fsize = None
        else:
            if data_object.is_root:
                root = data_object
            else:
                root = data_object.find_root()
            fsize = root.tree_size

        # Resolve field dependencies.
        fields_to_read, fields_to_generate = \
          fi.resolve_field_dependencies(fields, fcache=fcache,
                                        fsize=fsize)

        # Keep list of fields present before getting new ones.
        # We need to do this after trees have been setup since
        # that will add fields to the field cache in some cases.
        old_fields = list(fcache.keys())

        # Read in fields we need that are on disk.
        if fields_to_read:
            read_data = self._read_fields(
                storage_object, fields_to_read, **kwargs)
            fcache.update(read_data)

        # Generate all derived fields/aliases, but
        # only after dependencies have been generated.
        while len(fields_to_generate) > 0:
            field = fields_to_generate.pop(0)
            if fi[field].get("type") == "analysis":
                if field not in fields:
                    raise ArborAnalysisFieldNotGenerated(field, self.arbor)
                self._initialize_analysis_field(storage_object, field)
                continue
            deps = set(fi[field]["dependencies"])
            need = deps.difference(fcache)
            # have not created all dependencies yet, try again later
            if need:
                fields_to_generate.append(field)
            # all dependencies present, generate the field
            else:
                units = fi[field].get("units")
                ftype = fi[field]["type"]
                if ftype == "alias":
                    data = fcache[fi[field]["dependencies"][0]]
                elif ftype == "derived":
                    data = fi[field]["function"](fi[field], fcache)
                if hasattr(data, "units") and units is not None:
                    data.convert_to_units(units)
                fcache[field] = data

        self._store_fields(storage_object, set(old_fields).union(fields))
        return storage_object.field_data

class TreeFieldIO(FieldIO):
    """
    IO class for getting fields for a tree.
    """

    def _initialize_analysis_field(self, storage_object, name):
        if name in storage_object.field_data:
            return
        fi = self.arbor.field_info[name]
        units = fi.get('units', '')
        dtype = fi.get('dtype', self.default_dtype)
        value = fi.get('default', 0)
        data = np.full(storage_object.tree_size, value, dtype=dtype)
        if units:
            data = self.arbor.arr(data, units)
        storage_object.field_data[name] = data

    def _determine_field_storage(self, data_object):
        return data_object.find_root()

    def _read_fields(self, root_node, fields, dtypes=None,
                     root_only=False):
        """
        Read fields from disk for a single tree.
        """

        if dtypes is None:
            dtypes = {}
        my_dtypes = self._determine_dtypes(
            fields, override_dict=dtypes)

        if root_only:
            fsize = 1
        else:
            fsize = root_node.tree_size

        field_data = {}
        for field in fields:
            field_data[field] = \
              np.empty(fsize, dtype=my_dtypes[field])

        if root_only:
            my_nodes = [root_node]
        else:
            my_nodes = root_node._tree_nodes

        data_files = defaultdict(list)
        for node in my_nodes:
            data_files[node.data_file].append(node)

        for data_file, nodes in data_files.items():
            my_data = data_file._read_fields(fields, tree_nodes=nodes,
                                             dtypes=my_dtypes)
            for field in fields:
                for i, node in enumerate(nodes):
                    field_data[field][node.tree_id] = my_data[field][i]

        self._apply_units(fields, field_data)

        return field_data

class DefaultRootFieldIO(FieldIO):
    """
    Class for getting root fields from arbors that have no
    specialized storage for root fields.
    """

    def _initialize_analysis_field(self, storage_object, name):
        fi = self.arbor.field_info[name]
        default = fi['default']
        dtype   = fi['dtype']
        units   = fi['units']

        storage_object.field_data[name] = \
          self.arbor.arr(np.full(self.arbor.size, default, dtype=dtype), units)

    def _read_fields(self, storage_object, fields, dtypes=None,
                     root_only=True):
        if not fields:
            return

        if dtypes is None:
            dtypes = {}
        my_dtypes = self._determine_dtypes(
            fields, override_dict=dtypes)

        rvals = self.arbor._node_io_loop(
            self.arbor._node_io._read_fields,
            pbar="Reading root fields",
            fields=fields, dtypes=my_dtypes, root_only=True)

        field_data = \
          dict((field, np.concatenate([fvals[field] for fvals in rvals]))
               for field in fields)

        return field_data

class DataFile:
    """
    Base class for data files.

    This class allows us keep files open during i/o heavy operations
    and to keep things like caches of fields.
    """
    def __init__(self, filename):
        if not os.path.exists(filename):
            mylog.warning(
                f"Cannot find data file: {filename}. "
                 "Will not be able to load field data.")

        self.filename = filename
        self.fh = None

    def __repr__(self):
        return self.filename

    def open(self):
        raise NotImplementedError

    def close(self):
        if self.fh is not None:
            self.fh.close()
            self.fh = None


# A dict of arbor field generators.
arbor_fields = {}
arbor_fields['uid'] = lambda t: t.uid
# This will only be called for a root.
arbor_fields['desc_uid'] = lambda t: -1 if t.descendent is None \
  else t.descendent.uid

class CatalogDataFile(DataFile):
    """
    Base class for halo catalog files.
    """

    def __init__(self, filename, arbor):
        super().__init__(filename)
        self.arbor = weakref.proxy(arbor)
        self._parse_header()

    def _parse_header(self):
        """
        Load any relevant data from the file header.
        """
        raise NotImplementedError

    def _get_field_sources(self, fields):
        """
        Distinguish field sources.

        Distinguish fields to be read from disk, from the file header,
        and from arbor properties.
        """

        fi = self.arbor.field_info
        afields = [field for field in fields
                   if fi[field].get("source") == "arbor"]
        hfields = [field for field in fields
                   if fi[field].get("source") == "header"]
        rfields = set(fields).difference(hfields + afields)

        return afields, hfields, rfields

    def _create_field_arrays(self, fields, dtypes, size=None):
        """
        Initialize empty field arrays.
        """

        if size is None:
            field_data = dict((field, []) for field in fields)

        else:
            field_data = \
              dict((field, np.empty(size, dtype=dtypes[field]))
                   for field in fields)

        return field_data

    def _get_arbor_fields(self, afields, tree_nodes, dtypes):
        """
        Get fields from arbor/tree_node properties.
        """

        if not afields:
            return {}

        nt = len(tree_nodes)
        field_data = self._create_field_arrays(afields, dtypes, size=nt)

        for field in afields:
            for i in range(nt):
                field_data[field][i] = \
                  arbor_fields[field](tree_nodes[i])

        return field_data

    def _get_header_fields(self, hfields, tree_nodes, dtypes):
        """
        Get fields from file header.
        """

        if not hfields:
            return {}

        field_data = {}
        hfield_values = dict((field, getattr(self, field))
                             for field in hfields)
        nt = len(tree_nodes)
        for field in hfields:
            field_data[field] = hfield_values[field] * \
              np.ones(nt, dtypes[field])

        return field_data

    def _read_data_default(self, rfields, dtypes):
        """
        Read field data for all halos in the file.
        """
        raise NotImplementedError

    def _read_data_select(self, rfields, tree_nodes, dtypes):
        """
        Read field data for a given set of halos.
        """
        raise NotImplementedError

    def _read_fields(self, fields, tree_nodes=None, dtypes=None):
        """
        Read all requested fields from disk, header, or arbor properties.
        """
        if dtypes is None:
            dtypes = {}

        field_data = {}
        afields, hfields, rfields = self._get_field_sources(fields)

        if tree_nodes is None:
            field_data = self._read_data_default(
                fields, dtypes)

        else:
            # fields from the actual data
            field_data.update(
                self._read_data_select(
                    rfields, tree_nodes, dtypes))

            # fields from arbor-related info
            field_data.update(
                self._get_arbor_fields(
                    afields, tree_nodes, dtypes))

            # fields from the file header
            field_data.update(
                self._get_header_fields(
                    hfields, tree_nodes, dtypes))

        return field_data



"""
AnalysisPipeline operators



"""

class AnalysisOperation:
    """
    An analysis task performed by an AnalysisPipeline.

    This is an internal class that facilitates keeping track of a
    function, arguments, and keyword arguments that together represent a
    single operation in a pipeline.

    Parameters
    ----------
    function : callable
        A function that minimally accepts a
        :class:`~ytree.data_structures.tree_node.TreeNode` object. The
        function may also accept additional positional and keyword arguments.
    """
    def __init__(self, function, *args, always_do=False, **kwargs):
        self.function = function
        self.always_do = always_do
        self.args = args
        self.kwargs = kwargs

    def __call__(self, target):
        return self.function(target, *self.args, **self.kwargs)



"""
ytree config



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import configparser
import os

CONFIG_DIR = os.environ.get("XDG_CONFIG_HOME",
    os.path.join(os.path.expanduser("~"), ".config", "ytree"))

ytreecfg = configparser.ConfigParser()
ytreecfg.read(os.path.join(CONFIG_DIR, "ytreerc"))
if not ytreecfg.has_section("ytree"):
    ytreecfg.add_section("ytree")



"""
ytree imports



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.analysis.analysis_pipeline import AnalysisPipeline
from ytree.data_structures.load import \
    load
from ytree.data_structures.tree_node_selector import \
    TreeNodeSelector, \
    add_tree_node_selector
from ytree.visualization.tree_plot import \
    TreePlot

from ytree.utilities.parallel import \
    parallel_trees, \
    parallel_tree_nodes, \
    parallel_nodes

__version__ = '3.3.0'



from ytree.frontends.gadget4 import \
    Gadget4Arbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class Gadget4ArborSingleTest(TempDirTest, ArborTest):
    arbor_type = Gadget4Arbor
    test_filename = "gadget4/trees/trees.hdf5"
    groups = ("forest", "tree", "prog")
    num_data_files = 1
    tree_skip = 100



from ytree.frontends.consistent_trees import \
    ConsistentTreesGroupArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class ConsistentTreesGroupArborTest(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesGroupArbor
    test_filename = "tiny_ctrees/locations.dat"
    num_data_files = 8
    custom_vector_fields = (
        ("a_ratio", ("b_to_a", "c_to_a")),
        ("bunch_of_rs", ("r200c", "r200b", "r500c", "r500b")),
    )

    def test_data_files(self):
        self.arbor._plant_trees()
        ArborTest.test_data_files(self)



"""
miscellaneous utility tests



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
from ytree.data_structures.load import \
    load as ytree_load
from ytree.utilities.io import \
    f_text_block
from ytree.utilities.loading import \
    test_data_dir
from ytree.utilities.testing import \
    requires_file, \
    TempDirTest

R0 = os.path.join(test_data_dir, "rockstar/rockstar_halos/out_0.list")
R63 = os.path.join(test_data_dir, "rockstar/rockstar_halos/out_63.list")

@requires_file(R63)
def test_f_text_block():
    for block_size in [40, 32768]:
        lines = []
        locs = []
        f = open(R63, "r")
        for line, loc in f_text_block(f, block_size=block_size):
            lines.append(line)
            locs.append(loc)

        lines2 = []
        for loc in locs:
            f.seek(loc)
            line = f.readline()
            lines2.append(line.strip())

        for l1, l2 in zip(lines, lines2):
            assert l1 == l2

class MiscTest(TempDirTest):
    """
    Some miscellaneous tests in temporary directories
    """

    @requires_file(R0)
    def test_tree_field_clobber(self):
        """
        Test for an issue where the uid and desc_uid fields would be
        deleted if the tree was setup in the middle of getting fields.
        Fixed in PR #19: https://github.com/ytree-project/ytree/pull/19
        """
        a = ytree_load(R0)
        t = a[a['mass'].argmax()]
        t['prog', 'redshift']
        t.save_tree()

@requires_file(R0)
def test_field_access_without_tree_setup():
    """
    Test field access on non-root nodes of catalog arbors where
    setup_tree has not been called.
    Fixed in PR #20: https://github.com/ytree-project/ytree/pull/20
    """
    a = ytree_load(R0)
    t = a[a['mass'].argmax()]
    list(t.ancestors)[0]['desc_uid']



"""
tests for LHaloTree reader.


"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from numpy.testing import \
    assert_equal
import numpy as np
import os
import tempfile
import ytree
from ytree.utilities.loading import \
    test_data_dir
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest, \
    requires_file
from ytree.frontends.lhalotree import \
    LHaloTreeArbor, \
    utils as lhtutils


MMT = os.path.join(test_data_dir, 'lhalotree', 'trees_063.0')
SMT = os.path.join(test_data_dir, 'lhalotree', 'small_trees_063.0')
MMP = os.path.join(test_data_dir, 'lhalotree', 'millennium.param')
CTT = os.path.join(test_data_dir, 'consistent_trees', 'tree_0_0_0.dat')

class LHaloTreeArborTest(TempDirTest, ArborTest):
    arbor_type = LHaloTreeArbor
    test_filename = "lhalotree/trees_063.0"
    groups = ("forest", "tree", "prog")
    tree_skip = 100

@requires_file(MMT)
def make_small_example(ntrees_per_file=None):
    if ntrees_per_file is None:
        ntrees_per_file = np.array([2, 2], 'int')
    ntrees_before_file = np.cumsum(ntrees_per_file) - ntrees_per_file
    header_size0, nhalos_per_tree0 = lhtutils.read_header_default(MMT)
    data0 = lhtutils._read_from_mmap(
        MMT, header_size=header_size0,
        nhalo=np.sum(nhalos_per_tree0[:np.sum(ntrees_per_file)]))
    fname_base = os.path.splitext(SMT)[0]
    prev_halos = 0
    for i in range(len(ntrees_per_file)):
        ifile = f"{fname_base}.{i}"
        start_tree = ntrees_before_file[i]
        stop_tree = start_tree + ntrees_per_file[i]
        nhalos_per_tree1 = nhalos_per_tree0[start_tree:stop_tree]
        ntothalos1 = np.sum(nhalos_per_tree1)
        data1 = data0[prev_halos:(prev_halos + ntothalos1)]
        prev_halos += ntothalos1
        header_size1 = lhtutils.save_header_default(ifile, nhalos_per_tree1)
        lhtutils._save_to_mmap(ifile, data1, header_size=header_size1)


@requires_file(SMT)
def test_io_header_default():
    header_size0, nhalos_per_tree0 = lhtutils.read_header_default(SMT)
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        _header_size = lhtutils.save_header_default(tmp, nhalos_per_tree0)
        assert_equal(_header_size, header_size0)
        tmp.close()
        header_size1, nhalos_per_tree1 = lhtutils.read_header_default(tmp.name)
        os.remove(tmp.name)
    assert_equal(header_size1, header_size0)
    np.testing.assert_array_equal(nhalos_per_tree1, nhalos_per_tree0)


@requires_file(SMT)
def test_io_mmap():
    header_size0, nhalos_per_tree0 = lhtutils.read_header_default(SMT)
    data0 = lhtutils._read_from_mmap(SMT, header_size=header_size0)
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp.close()
        lhtutils._save_to_mmap(tmp.name, data0)
        data1 = lhtutils._read_from_mmap(tmp.name)
        os.remove(tmp.name)
    np.testing.assert_array_equal(data1, data0)


@requires_file(SMT)
def test_io_file():
    header_size0, nhalos_per_tree0 = lhtutils.read_header_default(SMT)
    data0 = lhtutils._read_from_file(SMT, header_size=header_size0)
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp.close()
        lhtutils._save_to_file(tmp.name, data0)
        data1 = lhtutils._read_from_file(tmp.name)
        os.remove(tmp.name)
    np.testing.assert_array_equal(data1, data0)


@requires_file(SMT)
def test_io_trees_default():
    header_size0, nhalos_per_tree0 = lhtutils.read_header_default(SMT)
    data0 = lhtutils.read_trees_default(SMT, header_size=header_size0)
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp.close()
        lhtutils.save_trees_default(tmp.name, data0)
        data1 = lhtutils.read_trees_default(tmp.name)
        os.remove(tmp.name)
    for k in data0.keys():
        assert (k in data1)
        np.testing.assert_array_equal(data1[k], data0[k])


@requires_file(SMT)
def test_LHaloTreeReader():
    reader = lhtutils.LHaloTreeReader(SMT)
    attr_list = ['filename', 'fileindex', 'filepattern',
                 'parameter_file', 'scale_factor_file', 'header_size',
                 'nhalos_per_tree', 'nhalos_before_tree', 'totnhalos',
                 'ntrees', 'item_dtype', 'raw_fields', 'add_fields', 'fields',
                 'treenum_arr', 'fobj', 'filenum', 'all_uids', 'all_desc_uids',
                 '_root_data',
                 'hubble_constant', 'omega_matter', 'omega_lambda', 'box_size',
                 'periodic', 'comoving', 'units_vel', 'units_len', 'units_mass',
                 'scale_factors', 'parameters']
    for a in attr_list:
        getattr(reader, a)
    # fd = reader.open()
    for i in range(reader.ntrees):
        reader.read_single_tree(i, validate=True)  # , fd=fd)
        reader.read_single_root(i, validate=True)  # , fd=fd)
        reader.read_single_halo(i, 0, validate=True)  # , fd=fd)
    # fd.close()


@requires_file(CTT)
def test_fail_load():
    assert (not LHaloTreeArbor._is_valid(CTT))

        
@requires_file(SMT)
@requires_file(MMP)
def test_load():
    a = ytree.load(SMT, parameter_file=MMP)
    print(a.size)
    print(a[0]["tree"])
    print(a[0]["tree", "mass"])
    print(a.trees)



"""
RockstarArbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from ytree.data_structures.fields import \
    FieldInfoContainer
from ytree.frontends.rockstar.misc import \
    Group

m_unit = "Msun"
p_unit = "unitary"
r_unit = "kpc"
v_unit = "km/s"

id_type = np.int64

class RockstarFieldInfo(FieldInfoContainer):
    alias_fields = (
        ("halo_id", "ID", None),
        ("desc_id", "DescID", None),
        ("mass", "Mvir", m_unit),
        ("virial_mass", "Mvir", m_unit),
        ("virial_radius", "Rvir", r_unit),
        ("scale_radius", "Rs", r_unit),
        ("velocity_dispersion", "Vrms", v_unit),
        ("position_x", "X", p_unit),
        ("position_y", "Y", p_unit),
        ("position_z", "Z", p_unit),
        ("velocity_x", "VX", v_unit),
        ("velocity_y", "VY", v_unit),
        ("velocity_z", "VZ", v_unit),
        ("angular_momentum_x", "JX", None),
        ("angular_momentum_y", "JY", None),
        ("angular_momentum_z", "JZ", None),
        ("spin_parameter", "Spin", None),
    )

    data_types = (
        ('ID', id_type),
        ('DescID', id_type),
        ('uid', id_type),
        ('desc_uid', id_type),
    )

def setup_field_groups():
    m = Group("masses")
    m.add_thing(lambda a: a.lower().startswith("m"))
    p = Group("positions")
    p.add_thing(lambda a: a.lower() in list("xyz"))
    v = Group("velocities")
    v.add_thing(lambda a: a.lower().startswith("v"))
    r = Group("radii")
    r.add_thing(lambda a: a.lower().startswith("r"))
    r.add_thing(lambda a: (len(a) > 1 and a.lower().startswith("x")))
    a = Group("angular")
    a.add_thing(lambda a: a.lower().startswith("j"))
    return [m, p, v, r, a]



from ytree.frontends.consistent_trees import \
    ConsistentTreesArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class ConsistentTreesArborTest(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesArbor
    test_filename = "consistent_trees/tree_0_0_0.dat"
    num_data_files = 1



import numpy as np
import types

from ytree.data_structures.fields import FieldContainer

_selection_types = ("forest", "tree", "prog")

class NodeContainer:
    """
    A container to hold a persistent collection of unrelated TreeNodes.

    This is a convenience object to hold a series of TreeNodes
    and provide simple field access similar to querying fields
    for an entire arbor. Field access for the container is cached
    for fast second-time access.

    Parameters
    ----------

    nodes : Iterable of TreeNodes
       A list, array, or generator of TreeNode objects.

    Examples
    --------

    >>> import ytree
    >>> a = ytree.load("tiny_ctrees/locations.dat")
    >>> first = a[0]
    >>> last = a[-1]
    >>> container = a.container([first, last])
    >>> for node in container:
    >>>     print (node["mass"])
    1.0231655e+12 Msun
    2.0633094e+11 Msun
    >>> print (container["mass"])
    [1.0231655e+12 2.0633094e+11] Msun
    >>> fn = a.save_arbor(trees=container)
    >>> a_new = ytree.load(fn)
    >>> print (a_new["mass"])
    [1.0231655e+12 2.0633094e+11] Msun

    >>> import ytree
    >>> a = ytree.load("tiny_ctrees/locations.dat")
    >>> a_slice = a.container(a[::8])
    >>> print (a.size, a_slice.size)
    32 4
    >>> print (a_slice["mass"])
    [1.0231655e+12 6.7683457e+11 2.9294966e+11 2.8014389e+11] Msun
    >>> my_tree = a_slice[0]
    >>> print (my_tree)
    TreeNode[1457223360]

    >>> import ytree
    >>> a = ytree.load("tiny_ctrees/locations.dat")
    >>> my_tree = a[0]
    >>> leaves = my_tree.get_leaf_nodes()
    >>> leaf_container = a.container(leaves)
    >>> print (leaf_container["mass"])
    [2.6503598e+08 4.2388490e+08 5.6964032e+08 5.5640288e+08 3.5769786e+08
     1.9870504e+08 6.3597126e+08] Msun
    >>> print (leaf_container.nodes)
    [TreeNode[1169524360], TreeNode[758740170], TreeNode[278238650],
     TreeNode[142383462], TreeNode[38885675], TreeNode[19602159],
     TreeNode[5057761]]

    """
    def __init__(self, nodes, arbor=None):
        self._nodes = nodes
        self.field_data = FieldContainer(arbor)

    @property
    def nodes(self):
        if isinstance(self._nodes, types.GeneratorType):
            self._nodes = list(self._nodes)
        return self._nodes

    def __len__(self):
        return len(self.nodes)

    @property
    def size(self):
        return self.__len__()

    def __iter__(self):
        for node in self.nodes:
            yield node

    def __getitem__(self, key):
        if isinstance(key, str):
            if key in _selection_types:
                raise SyntaxError("Argument must be a field or integer.")

            if key not in self.field_data:
                self.field_data[key] = self.field_data.arbor.arr(
                    [tree[key] for tree in self.nodes])
            return self.field_data[key]

        if isinstance(key, (int, np.integer, slice)):
            return self.nodes[key]

        else:
            raise ValueError(
                f"Unrecognized argument type: {key} ({type(key)}).")



"""
TreeFrogArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
import os
import re

from ytree.data_structures.arbor import \
    SegmentedArbor
from ytree.frontends.treefrog.fields import \
    TreeFrogFieldInfo
from ytree.frontends.treefrog.io import \
    TreeFrogDataFile, \
    TreeFrogRootFieldIO, \
    TreeFrogTreeFieldIO
from ytree.utilities.logger import \
    ytreeLogger as mylog

class TreeFrogArbor(SegmentedArbor):
    """
    Arbors loaded from consistent-trees data converted into HDF5.
    """

    _suffix = ".hdf5"
    _default_dtype = None
    _field_info_class = TreeFrogFieldInfo
    _root_field_io_class = TreeFrogRootFieldIO
    _tree_field_io_class = TreeFrogTreeFieldIO

    def __init__(self, filename):
        filename = self._determine_filename(filename)
        super().__init__(filename)

    def _determine_filename(self, filename):
        """
        Find the proper parameter filename if given a data file instead.
        """

        suffix = f".foreststats{self._suffix}"
        if filename.endswith(suffix):
            return filename

        match = re.search(rf"{self._suffix}\.\d+$", filename)
        forest_fn = filename[:match.start()] + suffix
        return forest_fn

    def _get_data_files(self):
        self.data_files = [TreeFrogDataFile(f"{self._prefix}{self._suffix}.{i}")
                           for i in range(self._nfiles)]

    @property
    def _prefix(self):
        suffix = f".foreststats{self._suffix}"
        return self.parameter_filename[:-len(suffix)]

    def _parse_parameter_file(self):
        with h5py.File(self.parameter_filename, mode="r") as f:
            self._nfiles = f["Header"].attrs["NFiles"]
            self._nsnaps = f["Header"].attrs["NSnaps"]
            self._size = f["ForestInfo"].attrs["NForests"]
            self._file_count = f["ForestInfo"]["NForestsPerFile"][()]

        if self._nfiles < 1:
            mylog.warning(f"Dataset {self.parameter_filename} has no data files.")
            return

        fn = f"{self._prefix}{self._suffix}.0"
        if not os.path.exists(fn):
            raise RuntimeError(f"Data file not found: {fn}.")

        with h5py.File(fn, mode="r") as f:
            self.hubble_constant = f["Header/Simulation"].attrs["h_val"]
            self.omega_matter = f["Header/Simulation"].attrs["Omega_m"]
            self.omega_lambda = f["Header/Simulation"].attrs["Omega_Lambda"]
            self.box_size = self.quan(f["Header/Simulation"].attrs["Period"], "Mpc/h")
            if self._nsnaps < 1:
                mylog.warning(f"Dataset {self.parameter_filename} has no snapshots.")
                return

            self.units = {}
            for attr in ["Length_unit_to_kpc",
                         "Mass_unit_to_solarmass",
                         "Velocity_unit_to_kms"]:
                self.units[attr] = f["Header/Units"].attrs[attr]

            field_list = list(f["Snap_000"].keys())
            self._scale_factors = \
              np.array([f[f"Snap_{i:03d}"].attrs["scalefactor"]
                        for i in range(self._nsnaps)])

        self.field_list = field_list
        self.field_info.update({field: {} for field in field_list})

        self.field_list.append("scale_factor")
        self.field_info["scale_factor"] = {"source": "arbor"}

    def _plant_trees(self):
        if self.is_planted or self._size == 0:
            return

        with h5py.File(self.parameter_filename, mode="r") as f:
            self._node_info['_tree_size'] = f["ForestInfo"]["ForestSizes"][()]

        ei = self._file_count.cumsum()
        si = ei - self._file_count
        trees = np.arange(self._size)
        fi = np.digitize(trees, ei)
        # number of the file each forest lives in
        self._node_info['_fi'] = fi
        # index within that file each forest is at
        self._node_info['_si'] = trees - si[fi]
        # This is slightly unsafe since the call to get_fields to get the uid field
        # will call _plant_trees, too. It will not get here because is_planted will
        # be True. As long as _fi and _si are initialized properly, it should be ok.
        self._node_info['uid'] = self["uid"]

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Should be an hdf5 file with a few key attributes.
        """
        fn = self._determine_filename(self, args[0])
        suffix = f".foreststats{self._suffix}"
        if not fn.endswith(suffix):
            return False
        if not h5py.is_hdf5(fn):
            return False

        info = \
          {"Header": ['FileSplittingCriteria', 'FilesSplitByForest',
                      'HaloCatalogBaseFileName', 'NFiles', 'NSnaps'],
          "ForestInfo": ['MaxForestFOFGroupID', 'MaxForestFOFGroupSize',
                         'MaxForestID', 'MaxForestSize', 'NForests']}

        with h5py.File(fn, mode='r') as f:
            for group, attrs in info.items():
                if group not in f:
                    return False
                for attr in attrs:
                    if attr not in f[group].attrs:
                        return False
        return True



"""
tests for parallel_nodes iterator



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os

from ytree.utilities.testing import ParallelTest, TempDirTest

script_path = os.path.dirname(__file__)

class ParallelNodesTest(TempDirTest, ParallelTest):
    test_script = os.path.join(script_path, "parallel/parallel_nodes.py")
    arg_sets = (
        ("all", "forest", 1, 0, 0, 0, 4),
        ("all", "tree",   1, 0, 0, 0, 4),
        ("all", "prog",   1, 0, 0, 0, 4),
        ("all", "forest", 0, 1, 0, 0, 4),
        ("all", "forest", 1, 0, 0, 0,  ), # sets save_every to None
        ("all", "forest", 1, 0, 0, 1, 4),
        ("all", "forest", 0, 1, 1, 0, 4),
    )

class NonRootParallelNodesTest(ParallelNodesTest):
    ncores = 2
    arg_sets = (
        ("nonroot", "tree", 1, 0, 0, 0,  ), # sets save_every to None
        ("nonroot", "tree", 0, 1, 0, 0,  ), # sets save_every to None
    )



from ytree.frontends.consistent_trees_hdf5 import \
    ConsistentTreesHDF5Arbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class ConsistentTreesHDF5ArborTest3(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesHDF5Arbor
    test_filename = "consistent_trees_hdf5/soa/forest_0.h5"
    num_data_files = 1
    tree_skip = 10000



"""
TreeNode class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import weakref

from ytree.data_structures.fields import \
    FieldContainer
from ytree.utilities.exceptions import \
    ArborUnsettableField

class TreeNode:
    """
    Class for objects stored in Arbors.

    Each TreeNode represents a halo in a tree.  A TreeNode knows
    its halo ID, the level in the tree, and its global ID in the
    Arbor that holds it.  It also has a list of its ancestors.
    Fields can be queried for it, its progenitor list, and the
    tree beneath.
    """

    _link = None

    def __init__(self, uid, arbor=None, root=False):
        """
        Initialize a TreeNode with at least its halo catalog ID and
        its level in the tree.
        """
        self.uid = uid
        self.arbor = weakref.proxy(arbor)
        if root:
            self.root = -1
            self.field_data = FieldContainer(arbor)
        else:
            self.root = None

    _tree_id = None # used by CatalogArbor
    @property
    def tree_id(self):
        """
        Return the index of this node in a list of all nodes in the tree.
        """
        if self.is_root:
            return 0
        elif self._link is not None:
            return self._link.tree_id
        else:
            return self._tree_id

    @tree_id.setter
    def tree_id(self, value):
        """
        Set the tree_id manually in CatalogArbors.
        """
        self._tree_id = value

    @property
    def is_root(self):
        """
        Is this node the last in the tree?
        """
        return self.root in [-1, self]

    def find_root(self):
        """
        Find the root node.
        """

        if self.is_root:
            return self

        root = self.root
        if root is not None:
            return root

        return self.walk_to_root()

    def walk_to_root(self):
        """
        Walk descendents until root.
        """

        my_node = self
        while not my_node.is_root:
            if my_node.descendent in (-1, None):
                break
            my_node = my_node.descendent
        return my_node

    def clear_fields(self):
        """
        If a root node, delete field data.
        If not root node, do nothing.
        """

        if not self.is_root:
            return
        self.field_data.clear()

    _descendent = None # used by CatalogArbor
    @property
    def descendent(self):
        """
        Return the descendent node.
        """

        if self.is_root:
            return None

        # set in CatalogArbor._plant_trees
        if self._descendent is not None:
            return self._descendent

        # conventional Arbor object
        desc_link = self._link.descendent
        if desc_link is None:
            return None
        return self.arbor._generate_tree_node(self.root, desc_link)

    _ancestors = None # used by CatalogArbor
    @property
    def ancestors(self):
        """
        Return a generator of ancestor nodes.
        """

        self.arbor._grow_tree(self)

        # conventional Arbor object
        if self._link is not None:
            for link in self._link.ancestors:
                yield self.arbor._generate_tree_node(self.root, link)
            return

        # If tree is not setup yet, the ancestor nodes will not have
        # root pointers yet.
        need_root = not self.arbor.is_setup(self)
        if need_root:
            root = self.walk_to_root()

        # set in CatalogArbor._plant_trees
        if self._ancestors is not None:
            for ancestor in self._ancestors:
                if need_root:
                    ancestor.root = root
                yield ancestor
            return
        return None

    _uids = None
    @property
    def uids(self):
        """
        Array of uids for all nodes in the tree.
        """
        if not self.is_root:
            return None
        if self._uids is None:
            self.arbor._build_attr("_uids", self)
        return self._uids

    _desc_uids = None
    @property
    def desc_uids(self):
        """
        Array of descendent uids for all nodes in the tree.
        """
        if not self.is_root:
            return None
        if self._desc_uids is None:
            self.arbor._build_attr("_desc_uids", self)
        return self._desc_uids

    _tree_size = None
    @property
    def tree_size(self):
        """
        Number of nodes in the tree.
        """
        if self._tree_size is not None:
            return self._tree_size
        if self.is_root:
            self.arbor._setup_tree(self)
            # pass back to the arbor to avoid calculating again
            self.arbor._store_node_info(self, '_tree_size')
        else:
            self._tree_size = len(list(self["tree"]))
        return self._tree_size

    _link_storage = None
    @property
    def _links(self):
        """
        Array of NodeLink objects with the ancestor/descendent structure.

        This is only used by conventional Arbor objects, i.e., not
        CatalogArbor objects.
        """
        if not self.is_root:
            return None
        if self._link_storage is None:
            self.arbor._build_attr("_link_storage", self)
        return self._link_storage

    def __setitem__(self, key, value):
        """
        Set analysis field value for this node.
        """

        fi = self.arbor.field_info[key]
        ftype = fi.get('type')
        if ftype not in ['analysis', 'analysis_saved']:
            raise ArborUnsettableField(key, self.arbor)

        vector_fieldname = fi.get("vector_fieldname", None)
        has_vector_field = vector_fieldname is not None

        if self.is_root:
            root = self
            tree_id = 0
            # if root, set the value in the arbor field storage
            self.arbor[key][self._arbor_index] = value
            if has_vector_field and vector_fieldname in self.arbor.field_data:
                del self.arbor.field_data[vector_fieldname]
        else:
            root = self.root
            tree_id = self.tree_id
        self.arbor._node_io.get_fields(self, fields=[key],
                                       root_only=False)
        data = root.field_data[key]
        data[tree_id] = value
        if has_vector_field and vector_fieldname in root.field_data:
            del root.field_data[vector_fieldname]

    def __getitem__(self, key):
        """
        Return field values or tree/prog generators.
        """
        return self.query(key)

    def query(self, key):
        """
        Return field values for this TreeNode, progenitor list, or tree.

        Parameters
        ----------
        key : string or tuple
            If a single string, it can be either a field to be queried or
            one of "tree" or "prog".  If a field, then return the value of
            the field for this TreeNode.  If "tree" or "prog", then return
            the list of TreeNodes in the tree or progenitor list.

            If a tuple, this can be either (string, string) or (string, int),
            where the first argument must be either "tree" or "prog".
            If second argument is a string, then return the field values
            for either the tree or the progenitor list.  If second argument
            is an int, then return the nth TreeNode in the tree or progenitor
            list list.

        Examples
        --------
        >>> # virial mass for this halo
        >>> print (my_tree["mvir"].to("Msun/h"))

        >>> # all TreeNodes in the progenitor list
        >>> print (my_tree["prog"])
        >>> # all TreeNodes in the entire tree
        >>> print (my_tree["tree"])

        >>> # virial masses for the progenitor list
        >>> print (my_tree["prog", "mvir"].to("Msun/h"))

        >>> # the 3rd TreeNode in the progenitor list
        >>> print (my_tree["prog", 2])

        Returns
        -------
        float, ndarray/unyt_array, TreeNode

        """
        arr_types = ("forest", "prog", "tree")
        if isinstance(key, tuple):
            if len(key) != 2:
                raise SyntaxError(
                    "Must be either 1 or 2 arguments.")
            ftype, field = key
            if ftype not in arr_types:
                raise SyntaxError(
                    f"First argument must be one of {str(arr_types)}.")
            if not isinstance(field, str):
                raise SyntaxError("Second argument must be a string.")

            self.arbor._node_io.get_fields(self, fields=[field], root_only=False)
            indices = getattr(self, f"_{ftype}_field_indices")

            data_object = self.find_root()
            return data_object.field_data[field][indices]

        else:
            if not isinstance(key, str):
                raise SyntaxError("Single argument must be a string.")

            # return the progenitor list or tree nodes in a list
            if key in arr_types:
                self.arbor._setup_tree(self)
                return getattr(self, f"_{key}_nodes")

            # return field value for this node
            self.arbor._node_io.get_fields(self, fields=[key],
                                           root_only=self.is_root)
            data_object = self.find_root()
            return data_object.field_data[key][self.tree_id]

    def __repr__(self):
        """
        Call me TreeNode.
        """
        return f"TreeNode[{self.uid}]"

    def get_node(self, selector, index):
        """
        Get a single TreeNode from a tree.

        Use this to get the nth TreeNode from a forest, tree, or
        progenitor list.

        For forest selection, the index value is absolute and refers
        to the entire forest in which the node belongs.

        For tree and prog selection, the index refers to the tree
        for which the calling TreeNode is the head.

        Parameters
        ----------
        selector : str ("forest", "tree", or "prog")
            The tree selector from which to get the TreeNode. This
            should be "forest", "tree", or "prog".
        index : int
            The index of the desired TreeNode in the forest, tree,
            or progenitor list.

        Returns
        -------
        node: :class:`~ytree.data_structures.tree_node.TreeNode`

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tiny_ctrees/locations.dat")
        >>> my_tree = a[0]
        >>> # get 6th TreeNode in the progenitor list
        >>> my_node = my_tree.get_node('prog', 5)

        """

        if selector in ("tree", "prog") and index == 0:
            return self

        self.arbor._setup_tree(self)
        self.arbor._grow_tree(self)
        indices = getattr(self, f"_{selector}_field_indices", None)
        if indices is None:
            raise RuntimeError("Bad selector.")

        my_link = self.root._links[indices][index]
        return self.arbor._generate_tree_node(self.root, my_link)

    def get_leaf_nodes(self, selector=None):
        """
        Get all leaf nodes from the tree of which this is the head.

        This returns a generator of all leaf nodes belonging to this
        tree. A leaf node is a node that has no ancestors.

        Parameters
        ----------
        selector : optional, str ("forest", "tree", or "prog")
            The tree selector from which leaf nodes will be found.
            If none given, this will be set to "forest" if the
            calling node is a root node and "tree" otherwise.

        Returns
        -------
        leaf_nodes : a generator of
            :class:`~ytree.data_structures.tree_node.TreeNode` objects.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tiny_ctrees/locations.dat")
        >>> my_tree = a[0]
        >>> for leaf in my_tree.get_leaf_nodes():
        ...     print (leaf["mass"])

        """

        if selector is None:
            if self.is_root:
                selector = "forest"
            else:
                selector = "tree"

        uids = self[selector, "uid"]
        desc_uids = self[selector, "desc_uid"]
        lids = np.where(~np.isin(uids, desc_uids))[0]
        for lid in lids:
            yield self.get_node(selector, lid)

    def get_root_nodes(self):
        """
        Get all root nodes from the forest to which this node belongs.

        This returns a generator of all root nodes in the forest. A root
        node is a node that has no descendents.

        Returns
        -------
        root_nodes : a generator of
            :class:`~ytree.data_structures.tree_node.TreeNode` objects.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("consistent_trees_hdf5/soa/forest.h5",
        ...                access="forest")
        >>> my_tree = a[0]
        >>> for root in my_tree.get_root_nodes():
        ...     print (root["mass"])

        """

        selector = "forest"
        desc_uids = self[selector, "desc_uid"]
        rids = np.where(desc_uids == -1)[0]
        for rid in rids:
            yield self.get_node(selector, rid)

    _ffi = slice(None)
    @property
    def _forest_field_indices(self):
        """
        Return default slice to select the whole forest.
        """
        return self._ffi

    @property
    def _forest_nodes(self):
        """
        An iterator over all TreeNodes in the forest.

        This is different from _tree_nodes in that we don't walk
        through the ancestors lists. We just yield every TreeNode
        there is.
        """

        self.arbor._grow_tree(self)
        root = self.root
        for link in root._links:
            yield self.arbor._generate_tree_node(self.root, link)

    @property
    def _tree_nodes(self):
        """
        An iterator over all TreeNodes in the tree beneath,
        starting with this TreeNode.

        For internal use only. Use the following instead:

        >>> for my_node in my_tree['tree']:
        ...     print (my_node)

        Examples
        --------

        >>> for my_node in my_tree._tree_nodes:
        ...     print (my_node)

        """

        self.arbor._grow_tree(self)
        yield self
        if self.ancestors is None:
            return
        for ancestor in self.ancestors:
            for a_node in ancestor._tree_nodes:
                yield a_node

    _tfi = None
    @property
    def _tree_field_indices(self):
        """
        Return the field array indices for all TreeNodes in
        the tree beneath, starting with this TreeNode.
        """

        if self._tfi is not None:
            return self._tfi

        self.arbor._grow_tree(self)
        self._tfi = np.array([node.tree_id for node in self._tree_nodes])
        return self._tfi

    @property
    def _prog_nodes(self):
        """
        An iterator over all TreeNodes in the progenitor list,
        starting with this TreeNode.

        For internal use only. Use the following instead:

        >>> for my_node in my_tree['prog']:
        ...     print (my_node)

        Examples
        --------

        >>> for my_node in my_tree._prog_nodes:
        ...     print (my_node)

        """

        self.arbor._grow_tree(self)
        my_node = self
        while my_node is not None:
            yield my_node
            ancestors = list(my_node.ancestors)
            if ancestors:
                my_node = my_node.arbor.selector(ancestors)
            else:
                my_node = None

    _pfi = None
    @property
    def _prog_field_indices(self):
        """
        Return the field array indices for all TreeNodes in
        the progenitor list, starting with this TreeNode.
        """

        if self._pfi is not None:
            return self._pfi

        self.arbor._grow_tree(self)
        self._pfi = np.array([node.tree_id for node in self._prog_nodes])
        return self._pfi

    def save_tree(self, filename=None, fields=None):
        r"""
        Save the tree to a file.

        The saved tree can be re-loaded as an arbor.

        Parameters
        ----------
        filename : optional, string
            Output file keyword.  Main header file will be named
            <filename>/<filename>.h5.
            Default: "tree_<uid>".
        fields : optional, list of strings
            The fields to be saved.  If not given, all
            fields will be saved.

        Returns
        -------
        filename : string
            The filename of the saved arbor.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("rockstar_halos/trees/tree_0_0_0.dat")
        >>> # save the first tree
        >>> fn = a[0].save_tree()
        >>> # reload it
        >>> a2 = ytree.load(fn)

        """

        if filename is None:
            filename = f"tree_{self.uid}"

        return self.arbor.save_arbor(
            filename=filename, fields=fields,
            trees=[self])



"""
tests for analysis pipeline



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal, assert_equal
import os
import ytree

from ytree.utilities.testing import \
    requires_file, \
    TempDirTest

TCL = "tiny_ctrees/locations.dat"

def set_test_field(node, output_dir=None):
    node["test_field"] = 2 * node["mass"]

def minimum_mass_filter(node, value):
    return node["mass"] > value

def set_always(node):
    node["test_always"] = 1

def my_recipe(ap):
    ap.add_operation(minimum_mass_filter, 3e11)
    ap.add_operation(set_test_field, output_dir="my_analysis")
    ap.add_operation(set_always, always_do=True)

class AnalysisPipelineTest(TempDirTest):
    test_filename = TCL

    def test_pipeline(self):

        a = ytree.load(self.test_filename)
        fn = a.save_arbor(trees=a[:8])

        a = ytree.load(fn)
        if "test_field" not in a.field_list:
            a.add_analysis_field("test_field", default=-1, units="Msun")
            a.add_analysis_field("test_always", dtype=int, default=0, units="")

        ap = ytree.AnalysisPipeline(output_dir="analysis")
        ap.add_recipe(my_recipe)

        trees = list(a[:])
        for tree in trees:
            for node in tree["forest"]:
                ap.process_target(node)

        assert os.path.exists("analysis/my_analysis")

        fn = a.save_arbor(trees=trees)
        a2 = ytree.load(fn)

        mf = a2["mass"] > 3e11
        assert_array_equal(a2["test_field"][mf], 2 * a2["mass"][mf])
        assert_array_equal(a2["test_field"][~mf], -np.ones((~mf).sum()))
        assert_equal(a2["test_always"].sum(), a2.size)
        
        for tree in a2:
            mf = tree["forest", "mass"] > 3e11
            assert_array_equal(tree["forest", "test_field"][mf],
                               2 * tree["forest", "mass"][mf])
            assert_array_equal(tree["forest", "test_field"][~mf],
                               -np.ones((~mf).sum()))

def increment_counter(node):
    if getattr(node, "count", None) is None:
        node.count = 0
    node["count"] = node.count
    node.count += 1

@requires_file(TCL)
def test_handoff_attrs():
    a = ytree.load(TCL)
    a.add_analysis_field("count", dtype=int, default=0, units="")

    ap = ytree.AnalysisPipeline(output_dir=".")
    ap.add_operation(increment_counter)

    nodes = list(a[:])
    for node in nodes:
        ap.process_target(node, handoff_attrs=["count"])

    for i, node in enumerate(nodes):
        # check value was set off of persistent attribute
        assert i == node["count"]
        # check the attribute as been deleted
        assert hasattr(node, "count") is False

def do_nothing(node):
    return


global_count = 0
def increment_global_counter():
    global global_count
    global_count += 1

@requires_file(TCL)
def test_preprocess():
    a = ytree.load(TCL)

    ap = ytree.AnalysisPipeline(output_dir=".")
    ap.add_operation(do_nothing,
                     preprocess_function=increment_global_counter)

    nodes = list(a[:])
    for node in nodes:
        ap.process_target(node)

    assert global_count == 1



"""
AHFArbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from ytree.data_structures.fields import \
    FieldInfoContainer

m_unit = "Msun/h"
p_unit = "kpc/h"
v_unit = "km/s"

id_type = np.int64

class AHFFieldInfo(FieldInfoContainer):
    known_fields = (
        ("Mhalo", m_unit),
        ("Mvir", m_unit),
        ("Xc", p_unit),
        ("Yc", p_unit),
        ("Zc", p_unit),
        ("VXc", v_unit),
        ("VYc", v_unit),
        ("VZc", v_unit),
        ("Rvir", p_unit),
        ("Rmax", p_unit),
        ("r2", p_unit),
        ("mbp_offset", p_unit),
        ("com_offset", p_unit),
        ("Vmax", v_unit),
        ("v_esc", v_unit),
        ("sigV", v_unit),
        ("Ekin", "Msun/h * (km/s)**2"),
        ("Epot", "Msun/h * (km/s)**2"),
        ("SurfP", "Msun/h * (km/s)**2"),
        ("Phi0", "(km/s)**2"),
    )

    alias_fields = (
        ("halo_id", "ID", ""),
        ("mass", "Mhalo", m_unit),
        ("mass", "Mvir", m_unit),
        ("virial_mass", "Mvir", m_unit),
        ("position_x", "Xc", p_unit),
        ("position_y", "Yc", p_unit),
        ("position_z", "Zc", p_unit),
        ("velocity_x", "VXc", v_unit),
        ("velocity_y", "VYc", v_unit),
        ("velocity_z", "VZc", v_unit),
        ("virial_radius", "Rvir", p_unit),
        ("velocity_dispersion", "sigV", v_unit),
        ("spin_parameter", "lambda", ""),
        ("kinetic_energy", "Ekin", "Msun/h * (km/s)**2"),
        ("potential_energy", "Epot", "Msun/h * (km/s)**2"),
    )

    data_types = (
        ('ID', id_type),
        ('desc_id', id_type),
        ('uid', id_type),
        ('desc_uid', id_type)
    )

class AHFCRMFieldInfo(AHFFieldInfo):
    alias_fields = (
        ("uid", "ID", ""),
        ("desc_uid", "desc_id", ""),
        ("mass", "Mhalo", m_unit),
        ("mass", "Mvir", m_unit),
        ("virial_mass", "Mvir", m_unit),
        ("position_x", "Xc", p_unit),
        ("position_y", "Yc", p_unit),
        ("position_z", "Zc", p_unit),
        ("velocity_x", "VXc", v_unit),
        ("velocity_y", "VYc", v_unit),
        ("velocity_z", "VZc", v_unit),
        ("virial_radius", "Rvir", p_unit),
        ("velocity_dispersion", "sigV", v_unit),
        ("spin_parameter", "lambda", ""),
        ("kinetic_energy", "Ekin", "Msun/h * (km/s)**2"),
        ("potential_energy", "Epot", "Msun/h * (km/s)**2"),
    )



"""
lhalotree frontend



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.frontends.lhalotree.arbor import \
    LHaloTreeArbor



"""
Calculate halo age for each halo in the dataset.

Halo ago is defined as the scale factor at which the
halo has reached half of its current mass.

NOTE: this script includes extra code to make it run within
the test suite. To run conventionally, remove the following
lines and return the code block in the middle to the proper
tabbing (i.e., 4 spaces to the left).

    from mpi4py import MPI
    comm = MPI.Comm.Get_parent()
    try:

    except BaseException:
        pass
    comm.Disconnect()

"""

import numpy as np
import yt
yt.enable_parallelism()
import ytree

def calc_a50(node):
    # main progenitor masses
    pmass = node["prog", "mass"]

    mh = 0.5 * node["mass"]
    m50 = pmass <= mh

    if not m50.any():
        th = node["scale_factor"]
    else:
        pscale = node["prog", "scale_factor"]
        # linearly interpolate
        i = np.where(m50)[0][0]
        slope = (pscale[i-1] - pscale[i]) / (pmass[i-1] - pmass[i])
        th = slope * (mh - pmass[i]) + pscale[i]

    node["a50"] = th


if __name__ == "__main__":
    # Remove the next three and final three lines to run conventionally.
    from mpi4py import MPI
    comm = MPI.Comm.Get_parent()
    try:

        a = ytree.load("tiny_ctrees/locations.dat")
        a.add_analysis_field("a50", "")

        ap = ytree.AnalysisPipeline()
        ap.add_operation(calc_a50)

        trees = list(a[:])
        for tree in ytree.parallel_nodes(trees, filename="halo_age"):
            yt.mylog.info(f"Processing {tree}.")
            ap.process_target(tree)

        if yt.is_root():
            a2 = ytree.load("halo_age/halo_age.h5")
            print (a2[0]["a50"])

    # Remove the next three lines to run conventionally.
    except BaseException:
        pass
    comm.Disconnect()



"""
RockstarArbor miscellany



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

class Group:
    def __init__(self, name=None):
        self.things = None
        if name is not None:
            self.add_thing(
                lambda a: name.lower() in a.lower())
        self.name = name
    def add_thing(self, thing):
        if self.things is None:
            self.things = []
        self.things.append(thing)
    def in_group(self, item):
        if self.things is None:
            return False
        for thing in self.things:
            try:
                if thing(item):
                    return True
            except Exception:
                continue
        return False



from ytree.frontends.ahf import \
    AHFArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class AHFNameConfigArborTest(TempDirTest, ArborTest):
    arbor_type = AHFArbor
    test_filename = "B25_N256_CDM_1LPT/AHF.B25_N256_CDM_1LPT.snap_055.parameter"
    load_kwargs = {"name_config":
                       {"ahf_prefix": "AHF.B25_N256_CDM_1LPT",
                        "mtree_prefix": "MTREE.B25_N256_CDM_1LPT.z39_adapt",
                        "mtree_suffix": ".AHF_croco"}}
    num_data_files = 2
    tree_skip = 100
